from mne_lsl.stream import StreamLSL
import numpy as np
import pytest
import time


@pytest.mark.parametrize("setup_lsl_player", ["example_stream"], indirect=True)
def test_lsl_stream_search(setup_lsl_player):
    from py_neuromodulation.stream import mnelsl_stream

    """ Test if the lsl stream search can find any streams after starting a player."""
    player = setup_lsl_player
    player.start_player(chunk_size=30, n_repeat=4)
    time.sleep(3)
    streams = mnelsl_stream.resolve_streams()
    
    assert len(streams) != 0, "No streams found in search"
    
    player.stop_player()


@pytest.mark.parametrize("setup_lsl_player", ["offline_test"], indirect=True)
def test_offline_lsl(
    setup_default_stream_fast_compute, setup_lsl_player, setup_default_data
):
    """ " Test the offline lsl player and stream, checking for sfreq, channel types and channel names."""

    raw, data, sfreq = setup_default_data
    player = setup_lsl_player
    player.start_player(chunk_size=30, n_repeat=1)
    time.sleep(3)

    data, stream = setup_default_stream_fast_compute

    features = stream.run(
        is_stream_lsl=True,
        stream_lsl_name="offline_test",
        out_dir="./test_data",
        experiment_name="test_offline_lsl",
    )
    # check sfreq
    if not raw.info["sfreq"] == stream.sfreq:
        raise ValueError(
            "Expected same sampling frequency in the stream and input file"
        )
    
    # since the PlayerLSL is started within a multprocessing.Process
    # the player object cannot be accessed
    # if not player._player.info["sfreq"] == stream.sfreq:
    #     raise ValueError("Expected same sampling frequency in the stream and player")

    # check types
    if not all(raw.get_channel_types() == stream.channels["type"]):
        raise ValueError("Channel types in data file are not matching the stream")

    # if not all(player.get_channel_types() == stream.channels["type"]):
    #    raise ValueError("Channel types in stream are not matching the player")

    # check names
    if not all(raw.ch_names == stream.channels["name"]):
        raise ValueError("Expected same channel names in the stream and input file")
    #if not all(player.info["ch_names"] == stream.channels["name"]):
    #    raise ValueError("Expected same channel names in the stream and player")

    if not features is not None:
        raise ValueError(
            "Expected features dataframe as stream.run() output but got None"
        )
    
    player.stop_player()


@pytest.mark.parametrize("setup_lsl_player", ["data_test"], indirect=True)
def test_lsl_data(setup_default_data, setup_lsl_player):
    """Check if 99% of the data is the same in the stream and the raw data."""
    import pandas as pd

    raw, data, sfreq = setup_default_data
    df_stream = pd.DataFrame()
    player = setup_lsl_player
    player.start_player(chunk_size=30, n_repeat=1)
    time.sleep(3)

    stream_player_check = StreamLSL(bufsize=2, name="data_test").connect()
    time.sleep(0.2)
    while stream_player_check.n_new_samples != 0:
        winsize = stream_player_check.n_new_samples / stream_player_check.info["sfreq"]
        data, ts = stream_player_check.get_data(winsize)
        df_stream = pd.concat([df_stream, pd.DataFrame(data)], axis=1)
        time.sleep(0.5)

    raw_sliced = pd.DataFrame(raw.get_data()).iloc[:, -df_stream.shape[1] :]
    df_stream.columns = range(len(df_stream.columns))
    raw_sliced_values = raw_sliced.values
    df_stream_values = df_stream.values
    same_values = np.zeros(raw_sliced.shape[1], dtype=bool)
    for i in range(raw_sliced.shape[1]):
        same_values[i] = np.any(
            np.all(raw_sliced_values == df_stream_values[:, i][:, np.newaxis], axis=0)
        )
    matching_percentage = np.sum(same_values) / len(same_values) * 100

    assert np.any(
        matching_percentage >= 99
    ), f"Expected same data in at least 99 percent of the samples but got {np.max(matching_percentage)} percent"

    player.stop_player()




from pathlib import Path
import pytest
import runpy


@pytest.mark.parametrize("example_filename", Path("examples").glob("*.py"))
def test_run_through_all_test(example_filename):
    print(f"Running {example_filename}")
    if "plot_8_cebra_example.py" not in example_filename.name:
        runpy.run_path(example_filename)



"""Test the nm_filter module."""

import numpy as np
import pytest
from py_neuromodulation.filter import MNEFilter


class TestMNEFilterData:
    """Test filter_data method of MNEFilter class."""

    @pytest.mark.parametrize(
        "filter_length",
        ["500ms", "999ms", "1999ms", "3999ms", "2s"],
    )
    def test_filter_length(self, filter_length) -> None:
        """Test different filter lengths."""
        f_ranges = [
            [13, 35],
        ]
        sfreq = 4000
        duration = 10
        times = np.linspace(0, duration, int(duration * sfreq))
        bandpass_filter = MNEFilter(
            f_ranges=f_ranges,
            sfreq=sfreq,
            filter_length=filter_length,
            l_trans_bandwidth=8,  # transition bandwidth needs to be adjusted for smaller filter length
            h_trans_bandwidth=8,
            verbose=None,
        )
        oscill_freqs = 50
        data = np.sin(2 * np.pi * times * oscill_freqs)
        data_filtered = bandpass_filter.filter_data(data)
        assert data_filtered.shape == (
            1,
            len(f_ranges),
            duration * sfreq,
        )

    def test_filter_1d(self) -> None:
        """Test filtering of 1d array with multiple frequency ranges."""
        f_ranges = [
            [4, 8],
            [8, 12],
            [13, 35],
            [60, 200],
            [200, 500],
        ]
        sfreq = 4000
        duration = 10
        times = np.linspace(0, duration, int(duration * sfreq))
        bandpass_filter = MNEFilter(
            f_ranges=f_ranges,
            sfreq=sfreq,
            filter_length="999ms",
            l_trans_bandwidth=4,
            h_trans_bandwidth=4,
            verbose=None,
        )
        oscill_freqs = 50
        data = np.sin(2 * np.pi * times * oscill_freqs)
        data_filtered = bandpass_filter.filter_data(data)
        assert data_filtered.shape == (
            1,
            len(f_ranges),
            duration * sfreq,
        )

    def test_filter_2d(self) -> None:
        """Test filtering of 2d array with multiple frequency ranges and multiple channels."""
        f_ranges = [
            [4, 8],
            [8, 12],
            [13, 35],
            [60, 200],
            [200, 500],
        ]
        sfreq = 4000
        duration = 10
        times = np.linspace(0, duration, int(duration * sfreq))
        bandpass_filter = MNEFilter(
            f_ranges=f_ranges,
            sfreq=sfreq,
            filter_length="999ms",
            l_trans_bandwidth=4,
            h_trans_bandwidth=4,
            verbose=None,
        )
        oscill_freqs = np.expand_dims(np.arange(10, 51, 10), axis=-1)
        data = np.sin(2 * np.pi * times * oscill_freqs)
        data_filtered = bandpass_filter.filter_data(data)
        assert data_filtered.shape == (
            oscill_freqs.shape[0],
            len(f_ranges),
            duration * sfreq,
        )



import pytest
import numpy as np
from pydantic import ValidationError

from py_neuromodulation.features import Bursts
from py_neuromodulation import NMSettings


def test_init_wrong_fband():
    settings = NMSettings.get_default()
    settings.bursts_settings.frequency_bands = ["wrong_band"]
    with pytest.raises(ValidationError):
        Bursts(settings, ["ch1", "ch2"], 1000)


def test_init_wrong_treshold():
    settings = NMSettings.get_default()
    settings.bursts_settings.threshold = -1
    with pytest.raises(ValidationError):
        Bursts(settings, ["ch1", "ch2"], 1000)


def test_init_wrong_timeduration():
    settings = NMSettings.get_default()
    settings.bursts_settings.time_duration_s = -1
    with pytest.raises(ValidationError):
        Bursts(settings, ["ch1", "ch2"], 1000)


def test_init_wrong_burst_feature_init():
    settings = NMSettings.get_default()
    settings.bursts_settings.burst_features.duration = -1
    with pytest.raises(ValidationError):
        Bursts(settings, ["ch1", "ch2"], 1000)


def test_bursting_duration():
    np.random.seed(0)
    settings = NMSettings.get_default()
    settings.features.bursts = True
    settings.postprocessing.feature_normalization = False

    sfreq = 1000
    NUM_CH = 1
    time_points_beta = np.arange(0, 1, 1 / sfreq)
    beta_freq = 18

    beta_wave = np.sin(2 * np.pi * beta_freq * time_points_beta)

    ch_names = ["ch0"]

    bursts = Bursts(settings, ch_names, sfreq)

    for _ in range(10):
        np.random.seed(0)
        data = np.random.random([NUM_CH, 1 * sfreq])
        f = bursts.calc_feature(data)

    np.random.seed(0)
    # the percentile of the hilbert transform of a continuous oscillation will be high
    # select better max amplitude
    bursts = Bursts(settings, ch_names, sfreq)
    f_burst = bursts.calc_feature(beta_wave + np.random.random([NUM_CH, 1 * sfreq]))

    assert (
        f["ch0_bursts_low_beta_amplitude_max"]
        < f_burst["ch0_bursts_low_beta_amplitude_max"]
    )

    assert (
        f["ch0_bursts_low_beta_duration_max"]
        < f_burst["ch0_bursts_low_beta_duration_max"]
    )



import numpy as np
from collections.abc import Sequence
from typing import TYPE_CHECKING
from pydantic import field_validator
from py_neuromodulation.utils.types import NMBaseModel, BoolSelector, NMFeature
from py_neuromodulation.utils.pydantic_extensions import (
    NMField,
    NMErrorList,
    create_validation_error,
)
from py_neuromodulation import logger

if TYPE_CHECKING:
    from py_neuromodulation.stream.settings import NMSettings
    from py_neuromodulation.filter import KalmanSettings


class BandpowerFeatures(BoolSelector):
    activity: bool = True
    mobility: bool = False
    complexity: bool = False


class BandPowerSettings(NMBaseModel):
    segment_lengths_ms: dict[str, int] = NMField(
        default={
            "theta": 1000,
            "alpha": 500,
            "low beta": 333,
            "high beta": 333,
            "low gamma": 100,
            "high gamma": 100,
            "HFA": 100,
        },
        custom_metadata={"field_type": "FrequencySegmentLength"},
    )
    bandpower_features: BandpowerFeatures = BandpowerFeatures()
    log_transform: bool = True
    kalman_filter: bool = False

    @field_validator("bandpower_features")
    @classmethod
    def bandpower_features_validator(cls, bandpower_features: BandpowerFeatures):
        if not len(bandpower_features.get_enabled()) > 0:
            raise create_validation_error(
                error_message="Set at least one bandpower_feature to True.",
                location=["bandpass_filter_settings", "bandpower_features"],
            )
        return bandpower_features

    def validate_fbands(self, settings: "NMSettings") -> NMErrorList:
        """_summary_

        :param settings: _description_
        :type settings: NMSettings
        :raises create_validation_error: _description_
        :raises create_validation_error: _description_
        :raises ValueError: _description_
        """
        errors: NMErrorList = NMErrorList()

        for fband_name, seg_length_fband in self.segment_lengths_ms.items():
            # Check that all frequency bands are defined in settings.frequency_ranges_hz
            if fband_name not in settings.frequency_ranges_hz:
                logger.warning(
                    f"Frequency band {fband_name} in bandpass_filter_settings.segment_lengths_ms"
                    " is not defined in settings.frequency_ranges_hz"
                )

            # Check that all segment lengths are smaller than settings.segment_length_features_ms
            if not seg_length_fband <= settings.segment_length_features_ms:
                errors.add_error(
                    f"segment length {seg_length_fband} needs to be smaller than "
                    f" settings['segment_length_features_ms'] = {settings.segment_length_features_ms}",
                    location=[
                        "bandpass_filter_settings",
                        "segment_lengths_ms",
                        fband_name,
                    ],
                )

        # Check that all frequency bands defined in settings.frequency_ranges_hz
        for fband_name in settings.frequency_ranges_hz.keys():
            if fband_name not in self.segment_lengths_ms:
                errors.add_error(
                    f"frequency range {fband_name} "
                    "needs to be defined in settings.bandpass_filter_settings.segment_lengths_ms",
                    location=[
                        "bandpass_filter_settings",
                        "segment_lengths_ms",
                        fband_name,
                    ],
                )

        return errors


class BandPower(NMFeature):
    def __init__(
        self,
        settings: "NMSettings",
        ch_names: Sequence[str],
        sfreq: float,
        use_kf: bool | None = None,
    ) -> None:
        settings.validate()

        self.bp_settings: BandPowerSettings = settings.bandpass_filter_settings
        self.kalman_filter_settings: KalmanSettings = settings.kalman_filter_settings
        self.sfreq = sfreq
        self.ch_names = ch_names
        self.KF_dict: dict = {}

        from py_neuromodulation.filter import MNEFilter

        self.bandpass_filter = MNEFilter(
            f_ranges=[
                tuple(frange) for frange in settings.frequency_ranges_hz.values()
            ],
            sfreq=self.sfreq,
            filter_length=self.sfreq - 1,
            verbose=False,
        )

        if use_kf or (use_kf is None and self.bp_settings.kalman_filter):
            self.init_KF("bandpass_activity")

        seglengths = self.bp_settings.segment_lengths_ms

        self.feature_params = []
        for ch_idx, ch_name in enumerate(self.ch_names):
            for f_band_idx, f_band in enumerate(settings.frequency_ranges_hz.keys()):
                seglength_ms = seglengths[f_band]
                seglen = int(np.floor(self.sfreq / 1000 * seglength_ms))
                for bp_feature in self.bp_settings.bandpower_features.get_enabled():
                    feature_name = "_".join([ch_name, "bandpass", bp_feature, f_band])
                    self.feature_params.append(
                        (
                            ch_idx,
                            f_band_idx,
                            seglen,
                            bp_feature,
                            feature_name,
                        )
                    )

    def init_KF(self, feature: str) -> None:
        from py_neuromodulation.filter import define_KF

        for f_band in self.kalman_filter_settings.frequency_bands:
            for channel in self.ch_names:
                self.KF_dict["_".join([channel, feature, f_band])] = define_KF(
                    self.kalman_filter_settings.Tp,
                    self.kalman_filter_settings.sigma_w,
                    self.kalman_filter_settings.sigma_v,
                )

    def update_KF(self, feature_calc: np.floating, KF_name: str) -> np.floating:
        if KF_name in self.KF_dict:
            self.KF_dict[KF_name].predict()
            self.KF_dict[KF_name].update(feature_calc)
            feature_calc = self.KF_dict[KF_name].x[0]
        return feature_calc

    def calc_feature(self, data: np.ndarray) -> dict:
        data = self.bandpass_filter.filter_data(data)

        feature_results = {}
        for (
            ch_idx,
            f_band_idx,
            seglen,
            bp_feature,
            feature_name,
        ) in self.feature_params:
            feature_results[feature_name] = self.calc_bp_feature(
                bp_feature, feature_name, data[ch_idx, f_band_idx, -seglen:]
            )

        return feature_results

    def calc_bp_feature(self, bp_feature, feature_name, data):
        match bp_feature:
            case "activity":
                feature_calc = np.var(data)
                if self.bp_settings.log_transform:
                    feature_calc = np.log10(feature_calc)
                if self.KF_dict:
                    feature_calc = self.update_KF(feature_calc, feature_name)
            case "mobility":
                feature_calc = np.sqrt(np.var(np.diff(data)) / np.var(data))
            case "complexity":
                feature_calc = self.calc_complexity(data)
            case _:
                raise ValueError(f"Unknown bandpower feature: {bp_feature}")

        return np.nan_to_num(feature_calc)

    @staticmethod
    def calc_complexity(data: np.ndarray) -> float:
        dat_deriv = np.diff(data)
        deriv_variance = np.var(dat_deriv)
        mobility = np.sqrt(deriv_variance / np.var(data))
        dat_deriv_2_var = np.var(np.diff(dat_deriv))
        deriv_mobility = np.sqrt(dat_deriv_2_var / deriv_variance)

        return deriv_mobility / mobility



import numpy as np

import py_neuromodulation as nm


def test_post_init_channels_change():
    """Test if post initialization of channels will also be ported to the feature computation."""

    data = np.random.random((10, 1000))
    fs = 1000

    stream = nm.Stream(sfreq=fs, data=data)

    # default channel names are "ch{i}"
    # every time the name changes, the "new_name" should also changes
    # this is however only done during initialization
    stream.channels["new_name"] = [f"new_ch_name_{i}" for i in range(10)]

    features = stream.run(
        out_dir="./test_data", experiment_name="test_post_init_nm_channels_change"
    )

    assert len([f for f in features.columns if "new_ch_name_0" in f]) != 0


def test_post_init_channels_used_channels_change_single_channel():
    """Test if post initialization of nm_settings will also be ported to the feature computation."""

    np.random.seed(0)
    data = np.random.random((3, 1000))
    sfreq = 1000
    stream = nm.Stream(sfreq=sfreq, data=data, sampling_rate_features_hz=11)
    stream.channels["used"] = 0
    stream.channels.loc[1, "used"] = 1

    features = stream.run(
        data,
        out_dir="./test_data",
        experiment_name="test_post_init_nm_channels_used_channels_change_single_channel",
    )

    chs_not_used = stream.channels[stream.channels["used"] == 0]["new_name"]

    assert (
        np.sum(
            [
                len([c for c in features.columns if c.startswith(ch_not_used)])
                for ch_not_used in chs_not_used
            ]
        )
        == 0
    )


def test_post_init_channels_used_channels_change_multiple_channel():
    """Test if post initialization of nm_settings will also be ported to the feature computation."""

    np.random.seed(0)
    data = np.random.random((3, 1000))
    sfreq = 1000
    stream = nm.Stream(sfreq=sfreq, data=data, sampling_rate_features_hz=11)
    stream.channels["used"] = 0
    stream.channels.loc[[0, 2], "used"] = 1

    features = stream.run(
        data,
        out_dir="./test_data",
        experiment_name="test_post_init_nm_channels_used_channels_change_multiple_channel",
    )

    chs_not_used = stream.channels[stream.channels["used"] == 0]["new_name"]

    assert (
        np.sum(
            [
                len([c for c in features.columns if c.startswith(ch_not_used)])
                for ch_not_used in chs_not_used
            ]
        )
        == 0
    )



from collections.abc import Iterable
import numpy as np

from typing import TYPE_CHECKING, Annotated, Literal
from pydantic import Field

from py_neuromodulation.utils.types import NMFeature, NMBaseModel
from py_neuromodulation.utils.pydantic_extensions import NMField

if TYPE_CHECKING:
    from py_neuromodulation import NMSettings


ListOfTwoStr = Annotated[list[str], Field(min_length=2, max_length=2)]


MNE_CONNECTIVITY_METHOD = Literal[
    "coh",
    "cohy",
    "imcoh",
    "cacoh",
    "mic",
    "mim",
    "plv",
    "ciplv",
    "ppc",
    "pli",
    "dpli",
    "wpli",
    "wpli2_debiased",
    "gc",
    "gc_tr",
]

MNE_CONNECTIVITY_MODE = Literal["multitaper", "fourier", "cwt_morlet"]


class MNEConnectivitySettings(NMBaseModel):
    method: MNE_CONNECTIVITY_METHOD = NMField(default="plv")
    mode: MNE_CONNECTIVITY_MODE = NMField(default="multitaper")
    channels: list[ListOfTwoStr] = []


class MNEConnectivity(NMFeature):
    def __init__(
        self,
        settings: "NMSettings",
        ch_names: Iterable[str],
        sfreq: float,
    ) -> None:
        self.settings = settings

        self.ch_names = ch_names
        self.sfreq = sfreq

        self.channels = settings.mne_connectivity_settings.channels

        # Params used by spectral_connectivity_epochs
        self.mode = settings.mne_connectivity_settings.mode
        self.method = settings.mne_connectivity_settings.method
        self.indices = ([], [])  # convert channel names to channel indices in data
        for con_idx in range(len(self.channels)):
            seed_name = self.channels[con_idx][0]
            target_name = self.channels[con_idx][1]
            seed_name_reref = [ch for ch in self.ch_names if ch.startswith(seed_name)][0]
            target_name_reref = [ch for ch in self.ch_names if ch.startswith(target_name)][0]
            self.indices[0].append(self.ch_names.index(seed_name_reref))
            self.indices[1].append(self.ch_names.index(target_name_reref))

        self.fbands = settings.frequency_ranges_hz
        self.fband_ranges: list = []
        self.result_keys = []

        self.prev_batch_shape: tuple = (-1, -1)  # sentinel value

    def calc_feature(self, data: np.ndarray) -> dict:
        from mne_connectivity import spectral_connectivity_epochs

        # n_jobs is here kept to 1, since setup of the multiprocessing Pool
        # takes longer than most batch computing sizes
        spec_out = spectral_connectivity_epochs(
            data=np.expand_dims(data, axis=0),  # add singleton epoch dimension
            sfreq=self.sfreq,
            method=self.method,
            mode=self.mode,
            indices=self.indices,
            verbose=False,
        )
        dat_conn: np.ndarray = spec_out.get_data()

        # Get frequency band ranges only for the first batch, it's already the same
        if len(self.fband_ranges) == 0:
            for fband_range in self.fbands.values():
                self.fband_ranges.append(
                    np.where(
                        (np.array(spec_out.freqs) >= fband_range[0])
                        & (np.array(spec_out.freqs) <= fband_range[1])
                    )[0]
                )

        feature_results = {}
        for con_idx in np.arange(dat_conn.shape[0]):
            for fband_idx, fband_name in enumerate(self.fbands):
                # TODO: Add support for max_fband and max_allfbands
                feature_results[
                    "_".join(
                        [
                            self.method,
                            self.channels[con_idx][0],  # seed channel name
                            "to",
                            self.channels[con_idx][1],  # target channel name
                            "mean_fband",
                            fband_name,
                        ]
                    )
                ] = np.mean(dat_conn[con_idx, self.fband_ranges[fband_idx]])

        # Store current experiment parameters to check if re-initialization is needed
        self.prev_batch_shape = data.shape

        return feature_results



import pytest
import numpy as np

from py_neuromodulation import processing, NMSettings


NORM_METHODS = processing.NormalizationSettings.list_normalization_methods()


def test_raw_normalization_init():
    """Test to check that raw normalizer can't be instantiated with a wrong normalization method."""

    settings = NMSettings.get_default()
    settings.raw_normalization_settings.normalization_method = "meann"  # type: ignore

    with pytest.raises(Exception):
        processing.RawNormalizer(sfreq=1000, settings=settings)


def test_feature_normalization_init():
    """Test to check that feature normalizer can't be instantiated with a wrong normalization method."""

    settings = NMSettings.get_default()
    settings.feature_normalization_settings.normalization_method = "meann"  # type: ignore

    with pytest.raises(Exception):
        processing.FeatureNormalizer(settings=settings)


def test_process_norm_features():
    """Test that FeatureNormalizer returns correct values (not nan or infinite).
    Also test that the previous data is stored correctly."""

    settings = NMSettings.get_default()
    settings.feature_normalization_settings.normalization_method = "mean"

    norm = processing.FeatureNormalizer(settings=settings)

    data = np.ones([1, 5])
    data_normed = norm.process(data)

    assert np.all(np.isfinite(data_normed))
    assert np.allclose(data, norm.previous)


def test_previous_size_FeatureNorm():
    """Test that previous batch data is clipped correctly when clip is enabled (default clip = 3)"""
    norm = processing.FeatureNormalizer(settings=NMSettings())

    num_features = 5

    for _ in range(150):
        np.random.seed(0)
        norm.process(np.random.random([1, num_features]))

    assert norm.previous.shape[0] < norm.num_samples_normalize


def test_zscore_feature_analysis():
    """Test that previous data is not clipped  when clip is set to False"""

    settings = NMSettings.get_default()
    settings.feature_normalization_settings.clip = False

    norm = processing.FeatureNormalizer(settings=settings)

    num_features = 5

    data_to_norm = np.zeros([1, num_features])
    data_normed = np.zeros([1, num_features])

    for _ in range(400):
        np.random.seed(0)
        data_to_norm = np.random.random([1, num_features])
        data_normed = norm.process(data_to_norm)

    expect_res = (
        norm.previous[:, 0].std() * data_normed[0, 0] + norm.previous[:, 0].mean()
    )

    assert pytest.approx(expect_res, 0.1) == data_to_norm[0, 0]


def test_zscore_raw_analysis():
    """Test that zscore is giving the expected results"""

    settings = NMSettings.get_default()
    settings.raw_normalization_settings.clip = False

    norm = processing.RawNormalizer(sfreq=10, settings=settings)

    num_samples = 100
    data_to_norm = np.zeros([1, num_samples])
    data_normed = np.zeros([1, num_samples])

    for _ in range(400):
        data_to_norm = np.random.random([1, num_samples])
        data_normed = norm.process(data_to_norm)

    expect_res = (
        norm.previous[:, 0].std() * data_normed[0, 0] + norm.previous[:, 0].mean()
    )

    assert np.allclose(expect_res, data_to_norm[0, 0], rtol=0.1, atol=0.1)


def test_all_norm_methods_raw():
    """Test that all raw normalization methods return correct values (not nan or infinite)"""

    settings = NMSettings.get_default()
    settings.raw_normalization_settings.clip = False

    for norm_method in NMSettings.list_normalization_methods():
        settings.raw_normalization_settings.normalization_method = norm_method

        norm = processing.RawNormalizer(sfreq=10, settings=settings)

        num_samples = 10

        data_to_norm = np.zeros([1, num_samples])
        data_normed = np.zeros([1, num_samples])

        for _ in range(10):
            np.random.seed(0)
            data_to_norm = np.random.random([1, num_samples])
            data_normed = norm.process(data_to_norm)

        assert np.all(np.isfinite(data_normed))


def test_all_norm_methods_feature():
    """Test that all feature normalization methods return correct values (not nan or infinite)"""

    settings = NMSettings.get_default()
    settings.feature_normalization_settings.clip = False

    for norm_method in NMSettings.list_normalization_methods():
        settings.feature_normalization_settings.normalization_method = norm_method

        norm = processing.FeatureNormalizer(settings=settings)

        num_samples = 10

        data_to_norm = np.zeros([1, num_samples])
        data_normed = np.zeros([1, num_samples])

        for i in range(10):
            np.random.seed(i)
            data_to_norm = np.random.random([1, num_samples])
            data_normed = norm.process(data_to_norm)

        assert np.all(np.isfinite(data_normed))



import numpy as np
from scipy import fft

from py_neuromodulation.filter import NotchFilter


def test_notch_filter_setup():
    # by Nyquist theorem, frequencies are computed up to half sfreq
    for sfreq in [150, 200, 500, 1000]:
        line_noise = 50

        notch_filter = NotchFilter(sfreq, line_noise)

        # the computed filter is saved in self.filter_bank

        data = np.random.random(sfreq)
        filtered_dat = notch_filter.process(data)

        Z_filtered = np.abs(fft.rfft(filtered_dat))
        Z_nonfiltered = np.abs(fft.rfft(data))
        freqs = fft.rfftfreq(sfreq, 1 / sfreq)
        idx = (np.abs(freqs - line_noise)).argmin()

        assert np.mean(Z_filtered[idx - 1 : idx + 1]) < np.mean(
            Z_nonfiltered[idx - 1 : idx + 1]
        ), (
            f"testing notch filter with sampling frequency {line_noise} failed"
            f" for comparison fft power vs no filtering"
        )



import pytest
import numpy as np

import py_neuromodulation as nm


@pytest.fixture
def setup_default_data():
    (
        RUN_NAME,
        PATH_RUN,
        PATH_BIDS,
        PATH_OUT,
        datatype,
    ) = nm.io.get_paths_example_data()

    (
        raw,
        data,
        sfreq,
        line_noise,
        coord_list,
        coord_names,
    ) = nm.io.read_BIDS_data(PATH_RUN=PATH_RUN)

    return raw, data, sfreq


@pytest.fixture
def setup_default_stream_fast_compute() -> tuple[np.ndarray, nm.Stream]:
    """This test function sets a data batch and automatic initialized M1 dataframe

    Args:
        PATH_PYNEUROMODULATION (string): Path to py_neuromodulation repository

    Returns:
        ieeg_batch (np.ndarray): (channels, samples)
        df_M1 (pd Dataframe): auto intialized table for rereferencing
        settings_wrapper (settings.py): settings.json
        fs (float): example sampling frequency
    """

    (
        RUN_NAME,
        PATH_RUN,
        PATH_BIDS,
        PATH_OUT,
        datatype,
    ) = nm.io.get_paths_example_data()

    (
        raw,
        data,
        sfreq,
        line_noise,
        coord_list,
        coord_names,
    ) = nm.io.read_BIDS_data(PATH_RUN=PATH_RUN)

    channels = nm.utils.set_channels(
        ch_names=raw.ch_names,
        ch_types=raw.get_channel_types(),
        reference="default",
        bads=raw.info["bads"],
        new_names="default",
        used_types=("ecog", "dbs", "seeg"),
        target_keywords=("MOV_RIGHT_CLEAN",),
    )

    settings = nm.NMSettings.get_default()
    settings.reset()
    settings.fooof_settings.aperiodic.exponent = True
    settings.fooof_settings.aperiodic.offset = True
    settings.features.fooof = True

    stream = nm.Stream(
        settings=settings,
        channels=channels,
        path_grids=None,
        verbose=True,
        sfreq=sfreq,
        line_noise=line_noise,
        coord_list=coord_list,
        coord_names=coord_names,
    )

    return data, stream


@pytest.fixture
def setup_lsl_player(request):
    """This test function sets a data batch and automatic initialized dataframe

    Args:
        PATH_PYNEUROMODULATION (string): Path to py_neuromodulation repository

    Returns:
        player (mne_lsl.player.PlayerLSL): LSL player object
    """

    name = request.param
    (
        RUN_NAME,
        PATH_RUN,
        PATH_BIDS,
        PATH_OUT,
        datatype,
    ) = nm.io.get_paths_example_data()
    (
        raw,
        data,
        sfreq,
        line_noise,
        coord_list,
        coord_names,
    ) = nm.io.read_BIDS_data(PATH_RUN=PATH_RUN)

    player = nm.stream.LSLOfflinePlayer(
        raw=raw,
        stream_name=name,
    )
    return player


@pytest.fixture
def setup_databatch():
    """This test function sets a data batch and automatic initialized M1 dataframe

    Args:
        PATH_PYNEUROMODULATION (string): Path to py_neuromodulation repository

    Returns:
        ieeg_batch (np.ndarray): (channels, samples)
        df_M1 (pd Dataframe): auto intialized table for rereferencing
        settings_wrapper (settings.py): settings.json
        fs (float): example sampling frequency
    """

    (
        RUN_NAME,
        PATH_RUN,
        PATH_BIDS,
        PATH_OUT,
        datatype,
    ) = nm.io.get_paths_example_data()

    (
        raw,
        data,
        sfreq,
        line_noise,
        coord_list,
        coord_names,
    ) = nm.io.read_BIDS_data(PATH_RUN=PATH_RUN)

    settings = nm.NMSettings.get_fast_compute()

    generator = nm.stream.RawDataGenerator(
        data,
        int(np.floor(sfreq)),
        settings.sampling_rate_features_hz,
        settings.segment_length_features_ms,
    )
    data_batch = next(generator, None)

    return [raw.ch_names, raw.get_channel_types(), raw.info["bads"], data_batch[1]]



"""Module for resampling."""

import numpy as np
from py_neuromodulation.utils.types import NMBaseModel, NMPreprocessor
from py_neuromodulation.utils.pydantic_extensions import NMField


class ResamplerSettings(NMBaseModel):
    resample_freq_hz: float = NMField(
        default=1000, gt=0, custom_metadata={"unit": "Hz"}
    )


class Resampler(NMPreprocessor):
    """Resample data.

    Parameters
    ----------
    sfreq : float
        Original sampling frequency.

    Attributes
    ----------
    up: float
        Factor to upsample by.
    """

    def __init__(
        self,
        sfreq: float,
        resample_freq_hz: float,
        **kwargs,
    ) -> None:
        self.settings = ResamplerSettings(resample_freq_hz=resample_freq_hz)

        ratio = float(resample_freq_hz / sfreq)
        if ratio == 1.0:
            self.up = 0.0
        else:
            self.up = ratio

    def process(self, data: np.ndarray) -> np.ndarray:
        """Resample raw data using mne.filter.resample.

        Parameters
        ----------
        data : np.ndarray
            Data to resample

        Returns
        -------
        np.ndarray
            Resampled data
        """
        if not self.up:
            return data

        from mne.filter import resample

        return resample(data.astype(np.float64), up=self.up, down=1.0)



# Expose feature settings
from py_neuromodulation.features.bispectra import BispectraSettings
from py_neuromodulation.features.coherence import CoherenceSettings
from py_neuromodulation.features.fooof import FooofSettings
from py_neuromodulation.features.mne_connectivity import MNEConnectivitySettings
from py_neuromodulation.features.nolds import NoldsSettings
from py_neuromodulation.features.sharpwaves import SharpwaveSettings
from py_neuromodulation.features.bursts import BurstsSettings
from py_neuromodulation.features.oscillatory import OscillatorySettings
from py_neuromodulation.features.bandpower import BandPowerSettings


# Expose feature classes
from py_neuromodulation.features.linelength import LineLength
from py_neuromodulation.features.hjorth_raw import Hjorth, Raw
from py_neuromodulation.features.bispectra import Bispectra
from py_neuromodulation.features.coherence import Coherence
from py_neuromodulation.features.fooof import FooofAnalyzer
from py_neuromodulation.features.mne_connectivity import MNEConnectivity
from py_neuromodulation.features.nolds import Nolds
from py_neuromodulation.features.sharpwaves import SharpwaveAnalyzer
from py_neuromodulation.features.bursts import Bursts
from py_neuromodulation.features.oscillatory import FFT, STFT, Welch
from py_neuromodulation.features.bandpower import BandPower

# Expose feature processor and custom feature functions
from py_neuromodulation.features.feature_processor import (
    FeatureProcessors,
    add_custom_feature,
    remove_custom_feature,
)



import numpy as np
from collections.abc import Sequence

from py_neuromodulation.features.feature_processor import NMFeature


class LineLength(NMFeature):
    def __init__(self, settings: dict, ch_names: Sequence[str], sfreq: float) -> None:
        self.ch_names = ch_names

    def calc_feature(self, data: np.ndarray) -> dict:

        line_length = np.mean(
            np.abs(np.diff(data, axis=-1)) / (data.shape[1] - 1), axis=-1
        )

        feature_results = {}
        for ch_idx, ch_name in enumerate(self.ch_names):
            feature_results[f"{ch_name}_LineLength"] = line_length[ch_idx]

        return feature_results



import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib import gridspec
import seaborn as sb
from pathlib import PurePath
from py_neuromodulation import logger, PYNM_DIR
from py_neuromodulation.utils.types import _PathLike


def plot_df_subjects(
    df,
    x_col="sub",
    y_col="performance_test",
    hue=None,
    title="channel specific performances",
    PATH_SAVE: _PathLike = "",
    figsize_tuple: tuple[float, float] = (5, 3),
):
    alpha_box = 0.4
    plt.figure(figsize=figsize_tuple, dpi=300)
    sb.boxplot(
        x=x_col,
        y=y_col,
        hue=hue,
        data=df,
        palette="viridis",
        showmeans=False,
        boxprops=dict(alpha=alpha_box),
        showcaps=True,
        showbox=True,
        showfliers=False,
        notch=False,
        whiskerprops={"linewidth": 2, "zorder": 10, "alpha": alpha_box},
        capprops={"alpha": alpha_box},
        medianprops=dict(linestyle="-", linewidth=5, color="gray", alpha=alpha_box),
    )

    ax = sb.stripplot(
        x=x_col,
        y=y_col,
        hue=hue,
        data=df,
        palette="viridis",
        dodge=True,
        s=5,
    )

    if hue is not None:
        n_hues = df[hue].nunique()

        handles, labels = ax.get_legend_handles_labels()
        plt.legend(
            handles[0:n_hues],
            labels[0:n_hues],
            bbox_to_anchor=(1.05, 1),
            loc=2,
            title=hue,
            borderaxespad=0.0,
        )
    plt.title(title)
    plt.ylabel(y_col)
    plt.xticks(rotation=90)
    if PATH_SAVE:
        plt.savefig(
            PATH_SAVE,
            bbox_inches="tight",
        )
    # plt.show()
    return plt.gca()


def plot_epoch(
    X_epoch: np.ndarray,
    y_epoch: np.ndarray,
    feature_names: list,
    z_score: bool | None = None,
    epoch_len: int = 4,
    sfreq: int = 10,
    str_title: str = "",
    str_label: str = "",
    ytick_labelsize: float | None = None,
):
    from scipy.stats import zscore

    if z_score is None:
        X_epoch = zscore(
            np.nan_to_num(np.nanmean(np.squeeze(X_epoch), axis=0)),
            axis=0,
            nan_policy="omit",
        ).T
    y_epoch = np.stack([np.array(y_epoch)])
    plt.figure(figsize=(6, 6))
    plt.subplot(211)
    plt.imshow(X_epoch, aspect="auto")
    plt.yticks(np.arange(0, len(feature_names), 1), feature_names, size=ytick_labelsize)
    plt.xticks(
        np.arange(0, X_epoch.shape[1], 1),
        np.round(np.arange(-epoch_len / 2, epoch_len / 2, 1 / sfreq), 2),
        rotation=90,
    )
    plt.gca().invert_yaxis()
    plt.xlabel("Time [s]")
    plt.title(str_title)

    plt.subplot(212)
    for i in range(y_epoch.shape[0]):
        plt.plot(y_epoch[i, :], color="black", alpha=0.4)
    plt.plot(
        y_epoch.mean(axis=0),
        color="black",
        alpha=1,
        linewidth=3.0,
        label="mean target",
    )
    plt.legend()
    plt.ylabel("Target")
    plt.title(str_label)
    plt.xticks(
        np.arange(0, X_epoch.shape[1], 1),
        np.round(np.arange(-epoch_len / 2, epoch_len / 2, 1 / sfreq), 2),
        rotation=90,
    )
    plt.xlabel("Time [s]")
    plt.tight_layout()


def reg_plot(
    x_col: str, y_col: str, data: pd.DataFrame, out_path_save: str | None = None
):
    from py_neuromodulation.analysis.stats import permutationTestSpearmansRho

    plt.figure(figsize=(4, 4), dpi=300)
    rho, p = permutationTestSpearmansRho(
        data[x_col],
        data[y_col],
        False,
        "R^2",
        5000,
    )
    sb.regplot(x=x_col, y=y_col, data=data)
    plt.title(f"{y_col}~{x_col} p={np.round(p, 2)} rho={np.round(rho, 2)}")

    if out_path_save is not None:
        plt.savefig(
            out_path_save,
            bbox_inches="tight",
        )


def plot_bar_performance_per_channel(
    ch_names,
    performances: dict,
    PATH_OUT: _PathLike,
    sub: str | None = None,
    save_str: str = "ch_comp_bar_plt.png",
    performance_metric: str = "Balanced Accuracy",
):
    """
    performances dict is output of ml_decode
    """
    plt.figure(figsize=(4, 3), dpi=300)
    if sub is None:
        sub = list(performances.keys())[0]
    plt.bar(
        np.arange(len(ch_names)),
        [performances[sub][p]["performance_test"] for p in performances[sub]],
    )
    plt.xticks(np.arange(len(ch_names)), ch_names, rotation=90)
    plt.xlabel("channels")
    plt.ylabel(performance_metric)
    plt.savefig(
        PurePath(PATH_OUT, save_str),
        bbox_inches="tight",
    )
    plt.close()


def plot_corr_matrix(
    feature: pd.DataFrame,
    feature_file: _PathLike = "",
    ch_name: str = "",
    feature_names: list[str] = [],
    show_plot=True,
    OUT_PATH: _PathLike = "",
    feature_name_plt="Features_corr_matr",
    save_plot: bool = False,
    save_plot_name: str = "",
    figsize: tuple[float, float] = (7, 7),
    title: str = "",
    cbar_vmin: float = -1,
    cbar_vmax: float = 1.0,
):
    # cut out channel name for each column
    if not ch_name:
        feature_col_name = [
            i[len(ch_name) + 1 :] for i in feature_names if ch_name in i
        ]
    else:
        feature_col_name = feature.columns

    plt.figure(figsize=figsize)
    if (
        len(feature_names) > 0
    ):  # Checking length to accomodate for tests passing a pandas Index
        corr = feature[feature_names].corr()
    else:
        corr = feature.corr()
    sb.heatmap(
        corr,
        xticklabels=feature_col_name,
        yticklabels=feature_col_name,
        vmin=cbar_vmin,
        vmax=cbar_vmax,
        cmap="viridis",
    )
    if not title:
        if ch_name:
            plt.title("Correlation matrix features channel: " + str(ch_name))
        else:
            plt.title("Correlation matrix")
    else:
        plt.title(title)

    # if len(feature_col_name) > 50:
    #    plt.xticks([])
    #    plt.yticks([])

    if save_plot:
        plt_path = (
            PurePath(OUT_PATH, save_plot_name)
            if save_plot_name
            else get_plt_path(
                OUT_PATH=OUT_PATH,
                feature_file=feature_file,
                ch_name=ch_name,
                str_plt_type=feature_name_plt,
                feature_name="_".join(feature_names),
            )
        )

        plt.savefig(plt_path, bbox_inches="tight")
        logger.info(f"Correlation matrix figure saved to {plt_path}")

    if not show_plot:
        plt.close()

    plt.tight_layout()

    return plt.gca()


def plot_feature_series_time(features) -> None:
    plt.imshow(features.T, aspect="auto")


def get_plt_path(
    OUT_PATH: _PathLike = "",
    feature_file: str = "",
    ch_name: str = "",
    str_plt_type: str = "",
    feature_name: str = "",
) -> _PathLike:
    """[summary]

    Parameters
    ----------
    OUT_PATH : str, optional
        folder of preprocessed runs, by default None
    feature_file : str, optional
        run_name, by default None
    ch_name : str, optional
        ch_name, by default None
    str_plt_type : str, optional
        type of plot, e.g. mov_avg_feature or corr_matr, by default None
    feature_name : str, optional
        e.g. bandpower, stft, sharpwave_prominence, by default None
    """
    filename = (
        str_plt_type
        + (("_ch_" + ch_name) if ch_name else "")
        + (("_" + feature_name) if feature_name else "")
        + ".png"
    )

    return PurePath(OUT_PATH, feature_file, filename)


def plot_epochs_avg(
    X_epoch: np.ndarray,
    y_epoch: np.ndarray,
    epoch_len: int,
    sfreq: int,
    feature_names: list[str] = [],
    feature_str_add: str = "",
    cut_ch_name_cols: bool = True,
    ch_name: str = "",
    label_name: str = "",
    normalize_data: bool = True,
    show_plot: bool = True,
    save: bool = False,
    OUT_PATH: _PathLike = "",
    feature_file: str = "",
    str_title: str = "Movement aligned features",
    ytick_labelsize=None,
    figsize_x: float = 8,
    figsize_y: float = 8,
) -> None:
    from scipy.stats import zscore

    # cut channel name of for axis + "_" for more dense plot
    if not feature_names:
        if cut_ch_name_cols and None not in (ch_name, feature_names):
            feature_names = [
                i[len(ch_name) + 1 :] for i in list(feature_names) if ch_name in i
            ]

    if normalize_data:
        X_epoch_mean = zscore(
            np.nanmean(np.squeeze(X_epoch), axis=0), axis=0, nan_policy="omit"
        ).T
    else:
        X_epoch_mean = np.nanmean(np.squeeze(X_epoch), axis=0).T

    if len(X_epoch_mean.shape) == 1:
        X_epoch_mean = np.expand_dims(X_epoch_mean, axis=0)

    plt.figure(figsize=(figsize_x, figsize_y))
    gs = gridspec.GridSpec(2, 1, height_ratios=[2.5, 1])
    plt.subplot(gs[0])
    plt.imshow(X_epoch_mean, aspect="auto")
    plt.yticks(np.arange(0, len(feature_names), 1), feature_names, size=ytick_labelsize)
    plt.xticks(
        np.arange(0, X_epoch.shape[1], int(X_epoch.shape[1] / 10)),
        np.round(np.arange(-epoch_len / 2, epoch_len / 2, epoch_len / 10), 2),
        rotation=90,
    )
    plt.xlabel("Time [s]")
    str_title = str_title
    if ch_name:
        str_title += f" channel: {ch_name}"
    plt.title(str_title)

    plt.subplot(gs[1])
    for i in range(y_epoch.shape[0]):
        plt.plot(y_epoch[i, :], color="black", alpha=0.4)
    plt.plot(
        y_epoch.mean(axis=0),
        color="black",
        alpha=1,
        linewidth=3.0,
        label="mean target",
    )
    plt.legend()
    plt.ylabel("Target")
    plt.title(label_name)
    plt.xticks(
        np.arange(0, X_epoch.shape[1], int(X_epoch.shape[1] / 10)),
        np.round(np.arange(-epoch_len / 2, epoch_len / 2, epoch_len / 10), 2),
        rotation=90,
    )
    plt.xlabel("Time [s]")
    plt.tight_layout()

    if save:
        plt_path = get_plt_path(
            OUT_PATH,
            feature_file,
            ch_name,
            str_plt_type="MOV_aligned_features",
            feature_name=feature_str_add,
        )
        plt.savefig(plt_path, bbox_inches="tight")
        logger.info(f"Feature epoch average figure saved to: {str(plt_path)}")
    if not show_plot:
        plt.close()


def plot_grid_elec_3d(
    cortex_grid: np.ndarray | None = None,
    ecog_strip: np.ndarray | None = None,
    grid_color: np.ndarray | None = None,
    strip_color: np.ndarray | None = None,
):
    ax = plt.axes(projection="3d")

    if cortex_grid is not None:
        grid_color = np.ones(cortex_grid.shape[0]) if grid_color is None else grid_color
        _ = ax.scatter3D(
            cortex_grid[:, 0],
            cortex_grid[:, 1],
            cortex_grid[:, 2],
            c=grid_color,
            s=300,
            alpha=0.8,
            cmap="viridis",
        )

    if ecog_strip is not None:
        strip_color = (
            np.ones(ecog_strip.shape[0]) if strip_color is None else strip_color
        )
        _ = ax.scatter(
            ecog_strip[:, 0],
            ecog_strip[:, 1],
            ecog_strip[:, 2],
            c=strip_color,
            s=500,  # Bug? Third argument is s, what is this value?
            alpha=0.8,
            cmap="gray",
            marker="o",
        )


def plot_all_features(
    df: pd.DataFrame,
    time_limit_low_s: float | None = None,
    time_limit_high_s: float | None = None,
    normalize: bool = True,
    ytick_labelsize: int = 4,
    clim_low: float | None = None,
    clim_high: float | None = None,
    save: bool = False,
    title="all_feature_plt.pdf",
    OUT_PATH: _PathLike = "",
    feature_file: str = "",
):
    from scipy.stats import zscore

    if time_limit_high_s is not None:
        df = df[df["time"] < time_limit_high_s * 1000]
    if time_limit_low_s is not None:
        df = df[df["time"] > time_limit_low_s * 1000]

    cols_plt = [c for c in df.columns if c != "time"]
    if normalize:
        data_plt = zscore(df[cols_plt], nan_policy="omit")
    else:
        data_plt = df[cols_plt]

    plt.figure()  # figsize=(7, 5), dpi=300
    plt.imshow(data_plt.T, aspect="auto")
    plt.xlabel("Time [s]")
    plt.ylabel("Feature Names")
    plt.yticks(np.arange(len(cols_plt)), cols_plt, size=ytick_labelsize)

    tick_num = np.arange(0, df.shape[0], int(df.shape[0] / 10))
    tick_labels = np.array(np.rint(df["time"].iloc[tick_num] / 1000), dtype=int)
    plt.xticks(tick_num, tick_labels)

    plt.title(f"Feature Plot {feature_file}")

    if clim_low is not None:
        plt.clim(vmin=clim_low)
    if clim_high is not None:
        plt.clim(vmax=clim_high)

    plt.colorbar()
    plt.tight_layout()

    if save:
        plt_path = PurePath(OUT_PATH, feature_file, title)
        plt.savefig(plt_path, bbox_inches="tight")


def read_plot_modules(
    PATH_PLOT: _PathLike = PYNM_DIR / "plots",
):
    """Read required .mat files for plotting

    Parameters
    ----------
    PATH_PLOT : regexp, optional
        path to plotting files, by default
    """
    from py_neuromodulation.utils.io import loadmat

    faces = loadmat(PurePath(PATH_PLOT, "faces.mat"))
    vertices = loadmat(PurePath(PATH_PLOT, "Vertices.mat"))
    grid = loadmat(PurePath(PATH_PLOT, "grid.mat"))["grid"]
    stn_surf = loadmat(PurePath(PATH_PLOT, "STN_surf.mat"))
    x_ver = stn_surf["vertices"][::2, 0]
    y_ver = stn_surf["vertices"][::2, 1]
    x_ecog = vertices["Vertices"][::1, 0]
    y_ecog = vertices["Vertices"][::1, 1]
    z_ecog = vertices["Vertices"][::1, 2]
    x_stn = stn_surf["vertices"][::1, 0]
    y_stn = stn_surf["vertices"][::1, 1]
    z_stn = stn_surf["vertices"][::1, 2]

    return (
        faces,
        vertices,
        grid,
        stn_surf,
        x_ver,
        y_ver,
        x_ecog,
        y_ecog,
        z_ecog,
        x_stn,
        y_stn,
        z_stn,
    )


class NM_Plot:
    def __init__(
        self,
        ecog_strip: np.ndarray | None = None,
        grid_cortex: np.ndarray | None = None,
        grid_subcortex: np.ndarray | None = None,
        sess_right: bool | None = False,
        proj_matrix_cortex: np.ndarray | None = None,
    ) -> None:
        self.grid_cortex = grid_cortex
        self.grid_subcortex = grid_subcortex
        self.ecog_strip = ecog_strip
        self.sess_right = sess_right
        self.proj_matrix_cortex = proj_matrix_cortex

        (
            self.faces,
            self.vertices,
            self.grid,
            self.stn_surf,
            self.x_ver,
            self.y_ver,
            self.x_ecog,
            self.y_ecog,
            self.z_ecog,
            self.x_stn,
            self.y_stn,
            self.z_stn,
        ) = read_plot_modules()

    def plot_grid_elec_3d(self) -> None:
        plot_grid_elec_3d(np.array(self.grid_cortex), np.array(self.ecog_strip))

    def plot_cortex(
        self,
        grid_cortex: np.ndarray | pd.DataFrame | None = None,
        grid_color: np.ndarray | None = None,
        ecog_strip: np.ndarray | None = None,
        strip_color: np.ndarray | None = None,
        sess_right: bool | None = None,
        save: bool = False,
        OUT_PATH: _PathLike = "",
        feature_file: str = "",
        feature_str_add: str = "",
        show_plot: bool = True,
        title: str = "Cortical grid",
        set_clim: bool = True,
        lower_clim: float = 0.5,
        upper_clim: float = 0.7,
        cbar_label: str = "Balanced Accuracy",
    ):
        """Plot MNI brain including selected MNI cortical projection grid + used strip ECoG electrodes
        Colorcoded by grid_color
        """

        if grid_cortex is None:
            if type(self.grid_cortex) is pd.DataFrame:
                grid_cortex = np.array(self.grid_cortex)
            else:
                grid_cortex = self.grid_cortex

        if ecog_strip is None:
            ecog_strip = self.ecog_strip

        if sess_right:
            grid_cortex[0, :] = grid_cortex[0, :] * -1  # type: ignore # Handled above

        fig, axes = plt.subplots(1, 1, facecolor=(1, 1, 1), figsize=(14, 9))
        axes.scatter(self.x_ecog, self.y_ecog, c="gray", s=0.01)
        axes.axes.set_aspect("equal", anchor="C")

        if grid_cortex is not None:
            grid_color = (
                np.ones(grid_cortex.shape[0]) if grid_color is None else grid_color
            )

            pos_ecog = axes.scatter(
                grid_cortex[:, 0],
                grid_cortex[:, 1],
                c=grid_color,
                s=150,
                alpha=0.8,
                cmap="viridis",
                label="grid points",
            )
            if set_clim:
                pos_ecog.set_clim(lower_clim, upper_clim)
        if ecog_strip is not None:
            strip_color = (
                np.ones(ecog_strip.shape[0]) if strip_color is None else strip_color
            )

            pos_ecog = axes.scatter(
                ecog_strip[:, 0],
                ecog_strip[:, 1],
                c=strip_color,
                s=400,
                alpha=0.8,
                cmap="viridis",
                marker="x",
                label="ecog electrode",
            )
        plt.axis("off")
        plt.legend()
        plt.title(title)
        if set_clim:
            pos_ecog.set_clim(lower_clim, upper_clim)
            cbar = fig.colorbar(pos_ecog)
            cbar.set_label(cbar_label)

        if save:
            plt_path = get_plt_path(
                OUT_PATH,
                feature_file,
                str_plt_type="PLOT_CORTEX",
                feature_name=feature_str_add,
            )
            plt.savefig(plt_path, bbox_inches="tight")
            logger.info(f"Feature epoch average figure saved to: {str(plt_path)}")
        if not show_plot:
            plt.close()



import numpy as np
from pathlib import PurePath, Path


# from numba import jit
import scipy.io as sio
import pandas as pd
import nibabel as nib
from matplotlib import pyplot as plt

from py_neuromodulation.analysis import reg_plot
from py_neuromodulation.utils.types import _PathLike
from py_neuromodulation import PYNM_DIR

LIST_STRUC_UNCONNECTED_GRIDPOINTS_HULL = [256, 385, 417, 447, 819, 914]
LIST_STRUC_UNCONNECTED_GRIDPOINTS_WHOLEBRAIN = [
    1,
    8,
    16,
    33,
    34,
    35,
    36,
    37,
    51,
    75,
    77,
    78,
    99,
    109,
    115,
    136,
    155,
    170,
    210,
    215,
    243,
    352,
    359,
    361,
    415,
    416,
    422,
    529,
    567,
    569,
    622,
    623,
    625,
    627,
    632,
    633,
    634,
    635,
    639,
    640,
    641,
    643,
    644,
    650,
    661,
    663,
    667,
    683,
    684,
    685,
    704,
    708,
    722,
    839,
    840,
    905,
    993,
    1011,
]


class ConnectivityChannelSelector:
    def __init__(
        self,
        whole_brain_connectome: bool = True,
        func_connectivity: bool = True,
    ) -> None:
        """ConnectivityChannelSelector

        Parameters
        ----------
        whole_brain_connectome : bool, optional
            if True a 1236 whole-brain point grid is chosen,
            if False, a 1025 point grid of the cortical hull is loaded,
            by default True
        func_connectivity : bool, optional
            if true, functional connectivity fMRI is loaded,
            if false structural dMRIby, default True
        """

        self.connectome_name = self._get_connectome_name(
            whole_brain_connectome, func_connectivity
        )

        self.whole_brain_connectome = whole_brain_connectome
        self.func_connectivity = func_connectivity

        self.PATH_CONN_DECODING = PYNM_DIR / "ConnectivityDecoding"

        if whole_brain_connectome:
            self.PATH_GRID = PurePath(
                self.PATH_CONN_DECODING,
                "mni_coords_whole_brain.mat",
            )
            self.grid = sio.loadmat(self.PATH_GRID)["downsample_ctx"]
            if not func_connectivity:
                # reduce the grid to only valid points that are not in LIST_STRUC_UNCONNECTED_GRIDPOINTS_WHOLEBRAIN
                self.grid = np.delete(
                    self.grid,
                    LIST_STRUC_UNCONNECTED_GRIDPOINTS_WHOLEBRAIN,
                    axis=0,
                )
        else:
            self.PATH_GRID = PurePath(
                self.PATH_CONN_DECODING,
                "mni_coords_cortical_surface.mat",
            )
            self.grid = sio.loadmat(self.PATH_GRID)["downsample_ctx"]
            if not func_connectivity:
                # reduce the grid to only valid points that are not in LIST_STRUC_UNCONNECTED_GRIDPOINTS_HULL
                self.grid = np.delete(
                    self.grid, LIST_STRUC_UNCONNECTED_GRIDPOINTS_HULL, axis=0
                )

        if func_connectivity:
            self.RMAP_arr = nib.load(
                PurePath(self.PATH_CONN_DECODING, "RMAP_func_all.nii")
            ).get_fdata()
        else:
            self.RMAP_arr = nib.load(
                PurePath(self.PATH_CONN_DECODING, "RMAP_struc.nii")
            ).get_fdata()

    def _get_connectome_name(self, whole_brain_connectome: str, func_connectivity: str):
        connectome_name = "connectome_"
        if whole_brain_connectome:
            connectome_name += "whole_brain_"
        else:
            connectome_name += "hull_"
        if func_connectivity:
            connectome_name += "func"
        else:
            connectome_name += "struc"
        return connectome_name

    def get_available_connectomes(self) -> list:
        """Return list of saved connectomes in the
        package folder/ConnectivityDecoding/connectome_folder/ folder.

        Returns
        -------
        list_connectomes: list
        """
        return list(Path(self.PATH_CONN_DECODING, "connectome_folder").iterdir())

    def plot_grid(self) -> None:
        """Plot the loaded template grid that passed coordinates are matched to."""

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.scatter(self.grid[:, 0], self.grid[:, 1], self.grid[:, 2], s=50, alpha=0.2)
        plt.show()

    def get_closest_node(self, coord: list | np.ndarray) -> tuple[list, list]:
        """Given a list or np.array of coordinates, return the closest nodes in the
        grid and their indices.

        Parameters
        ----------
        coord : np.ndarray
            MNI coordinates with shape (num_channels, 3)

        Returns
        -------
        Tuple[list, list]
            Grid coordinates, grid indices
        """

        idx_ = []
        for c in coord:
            dist = np.linalg.norm(self.grid - c, axis=1)
            idx_.append(np.argmin(dist))

        return [self.grid[idx] for idx in idx_], idx_

    def get_rmap_correlations(
        self, fps: list | np.ndarray, RMAP_use: np.ndarray | None = None
    ) -> list:
        """Calculate correlations of passed fingerprints with the RMAP

        Parameters
        ----------
        fps : Union[list, np.array]
            List of fingerprints
        RMAP_use : np.ndarray, optional
            Passed RMAP, by default None

        Returns
        -------
        List
            correlation values
        """

        RMAP_ = self.RMAP_arr if RMAP_use is None else RMAP_use
        RMAP_ = RMAP_.flatten()
        corrs = []
        for fp in fps:
            corrs.append(np.corrcoef(RMAP_, fp.flatten())[0][1])
        return corrs

    def load_connectome(
        self,
        whole_brain_connectome: bool | None = None,
        func_connectivity: bool | None = None,
    ) -> None:
        """Load connectome, if not available download connectome from
        Zenodo.

        Parameters
        ----------
        whole_brain_connectome : bool, optional
            if true whole brain connectome
            if false cortical hull grid connectome, by default None
        func_connectivity : bool, optional
            if true fMRI if false dMRI, by default None
        """

        if whole_brain_connectome is not None:
            self.whole_brain_connectome = whole_brain_connectome
        if func_connectivity is not None:
            self.func_connectivity = func_connectivity

        self.connectome_name = self._get_connectome_name(
            self.whole_brain_connectome, self.func_connectivity
        )

        PATH_CONNECTOME = Path(
            self.PATH_CONN_DECODING,
            "connectome_folder",
            self.connectome_name + ".mat",
        )

        if not PATH_CONNECTOME.exists():
            user_input = input(
                "Do you want to download the connectome? (yes/no): "
            ).lower()
            if user_input == "yes":
                self._download_connectome()
            elif user_input == "no":
                print("Connectome missing, has to be downloaded")

        self.connectome = sio.loadmat(PATH_CONNECTOME)

    def get_grid_fingerprints(self, grid_idx: list | np.ndarray) -> list:
        return [self.connectome[str(grid_idx)] for grid_idx in grid_idx]

    def download_connectome(
        self,
    ):
        
        from urllib.request import urlretrieve

        # download the connectome from the Zenodo API
        print("Downloading the connectome...")

        record_id = "10804702"
        file_name = self.connectome_name

        filepath = Path(self.PATH_CONN_DECODING, "connectome_folder")
        filepath.mkdir(parents=True, exist_ok=True)

        urlretrieve(
            f"https://zenodo.org/api/records/{record_id}/files/{file_name}/content",
            filepath / f"{self.connectome_name}.mat",
        )


class RMAPCross_Val_ChannelSelector:
    def __init__(self) -> None:
        pass

    def load_fingerprint(self, path_nii) -> None:
        """Return Nifti fingerprint"""
        epi_img = nib.load(path_nii)
        self.affine = epi_img.affine
        fp = epi_img.get_fdata()
        return fp

    def load_all_fingerprints(self, path_dir: str, cond_str: str = "_AvgR_Fz.nii"):
        if cond_str is not None:
            l_fps = list(filter(lambda k: cond_str in str(k), Path(path_dir).iterdir()))
        else:
            l_fps = list(Path(path_dir).iterdir())

        return l_fps, [self.load_fingerprint(PurePath(path_dir, f)) for f in l_fps]

    def get_fingerprints_from_path_with_cond(
        self,
        path_dir: _PathLike,
        str_to_omit: str = "",
        str_to_keep: str = "",
        keep: bool = True,
    ) -> tuple[list, list]:
        l_fps = []

        if keep and str_to_keep:
            l_fps = list(
                filter(
                    lambda k: "_AvgR_Fz.nii" in str(k) and str_to_keep in str(k),
                    Path(path_dir).iterdir(),
                )
            )

        elif not keep and str_to_omit:
            l_fps = list(
                filter(
                    lambda k: "_AvgR_Fz.nii" in str(k) and str_to_omit not in str(k),
                    Path(path_dir).iterdir(),
                )
            )

        return l_fps, [self.load_fingerprint(PurePath(path_dir, f)) for f in l_fps]

    @staticmethod
    def save_Nii(
        fp: np.ndarray,
        affine: np.ndarray,
        name: str = "img.nii",
        reshape: bool = True,
    ):
        if reshape:
            fp = np.reshape(fp, (91, 109, 91), order="C")

        img = nib.nifti1.Nifti1Image(fp, affine=affine)

        nib.save(img, name)

    def get_RMAP(self, X: np.ndarray, y: np.ndarray):
        # faster than calculate_RMap_numba
        # https://stackoverflow.com/questions/71252740/correlating-an-array-row-wise-with-a-vector/71253141#71253141

        r = (
            len(y) * np.sum(X * y[None, :], axis=-1) - (np.sum(X, axis=-1) * np.sum(y))
        ) / (
            np.sqrt(
                (len(y) * np.sum(X**2, axis=-1) - np.sum(X, axis=-1) ** 2)
                * (len(y) * np.sum(y**2) - np.sum(y) ** 2)
            )
        )
        return r

    @staticmethod
    # @jit(nopython=True)
    def calculate_RMap_numba(fp, performances):
        # The RMap also needs performances; for every fingerprint / channel
        # Save the corresponding performance
        # for every voxel; correlate it with performances

        arr = fp[0].flatten()
        NUM_VOXELS = arr.shape[0]
        LEN_FPS = len(fp)
        fp_arr = np.empty((NUM_VOXELS, LEN_FPS))
        for fp_idx, fp_ in enumerate(fp):
            fp_arr[:, fp_idx] = fp_.flatten()

        RMAP = np.zeros(NUM_VOXELS)
        for voxel in range(NUM_VOXELS):
            corr_val = np.corrcoef(fp_arr[voxel, :], performances)[0][1]

            RMAP[voxel] = corr_val

        return RMAP

    @staticmethod
    # @jit(nopython=True)
    def get_corr_numba(fp, fp_test):
        val = np.corrcoef(fp_test, fp)[0][1]
        return val

    def leave_one_ch_out_cv(self, l_fps_names: list, l_fps_dat: list, l_per: list):
        # l_fps_dat is not flattened

        per_left_out = []
        per_predict = []

        for idx_left_out, f_left_out in enumerate(l_fps_names):
            # print(idx_left_out)
            l_cv = l_fps_dat.copy()
            per_cv = l_per.copy()

            l_cv.pop(idx_left_out)
            per_cv.pop(idx_left_out)

            conn_arr = []
            for f in l_cv:
                conn_arr.append(f.flatten())
            conn_arr = np.array(conn_arr)

            rmap_cv = np.nan_to_num(self.get_RMAP(conn_arr.T, np.array(per_cv)))

            per_predict.append(
                np.nan_to_num(
                    self.get_corr_numba(rmap_cv, l_fps_dat[idx_left_out].flatten())
                )
            )
            per_left_out.append(l_per[idx_left_out])
        return per_left_out, per_predict

    def leave_one_sub_out_cv(
        self, l_fps_names: list, l_fps_dat: list, l_per: list, sub_list: list
    ):
        # l_fps_dat assume non flatted arrays
        # each fp including the sub_list string will be iteratively removed for test set

        per_predict = []
        per_left_out = []

        for subject_test in sub_list:
            # print(subject_test)
            idx_test = [idx for idx, f in enumerate(l_fps_names) if subject_test in f]
            idx_train = [
                idx for idx, f in enumerate(l_fps_names) if subject_test not in f
            ]
            l_cv = list(np.array(l_fps_dat)[idx_train])
            per_cv = list(np.array(l_per)[idx_train])

            conn_arr = []
            for f in l_cv:
                conn_arr.append(f.flatten())
            conn_arr = np.array(conn_arr)
            rmap_cv = np.nan_to_num(self.get_RMAP(conn_arr.T, np.array(per_cv)))

            for idx in idx_test:
                per_predict.append(
                    np.nan_to_num(
                        self.get_corr_numba(rmap_cv, l_fps_dat[idx].flatten())
                    )
                )
                per_left_out.append(l_per[idx])
        return per_left_out, per_predict

    def get_highest_corr_sub_ch(
        self,
        cohort_test: str,
        sub_test: str,
        ch_test: str,
        cohorts_train: dict,
        path_dir: str = r"C:\Users\ICN_admin\OneDrive - Charité - Universitätsmedizin Berlin\Connectomics\DecodingToolbox_BerlinPittsburgh_Beijing\functional_connectivity",
    ):
        fp_test = self.get_fingerprints_from_path_with_cond(
            path_dir=path_dir,
            str_to_keep=f"{cohort_test}_{sub_test}_ROI_{ch_test}",
            keep=True,
        )[1][
            0
        ].flatten()  # index 1 for getting the array, 0 for the list fp that was found

        fp_pairs = []

        for cohort in cohorts_train.keys():
            for sub in cohorts_train[cohort]:
                fps_name, fps = self.get_fingerprints_from_path_with_cond(
                    path_dir=path_dir,
                    str_to_keep=f"{cohort}_{sub}_ROI",
                    keep=True,
                )

                for fp, fp_name in zip(fps, fps_name):
                    ch = fp_name[fp_name.find("ROI") + 4 : fp_name.find("func") - 1]
                    corr_val = self.get_corr_numba(fp_test, fp)
                    fp_pairs.append([cohort, sub, ch, corr_val])

        idx_max = np.argmax(np.array(fp_pairs)[:, 3])
        return fp_pairs[idx_max][0:3]

    def plot_performance_prediction_correlation(
        per_left_out, per_predict, out_path_save: str | None = None
    ):
        df_plt_corr = pd.DataFrame()
        df_plt_corr["test_performance"] = per_left_out
        df_plt_corr["struct_conn_predict"] = (
            per_predict  # change "struct" with "funct" for functional connectivity
        )

        reg_plot(
            x_col="test_performance",
            y_col="struct_conn_predict",
            data=df_plt_corr,
            out_path_save=out_path_save,
        )



import logging
import importlib.metadata
from datetime import datetime
from pathlib import Path
import os

from fastapi import (
    FastAPI,
    HTTPException,
    Query,
    WebSocket,
)
from fastapi.staticfiles import StaticFiles
from fastapi.middleware.cors import CORSMiddleware
from pydantic import ValidationError

from . import app_pynm
from .app_socket import WebsocketManager
from .app_utils import is_hidden, get_quick_access
import pandas as pd

from py_neuromodulation import PYNM_DIR, NMSettings
from py_neuromodulation.utils.types import FileInfo

# TODO: maybe pull this list from the MNE package?
ALLOWED_EXTENSIONS = [".npy", ".vhdr", ".fif", ".edf", ".bdf"]

DEV_SERVER_PORT = 54321


class PyNMBackend(FastAPI):
    def __init__(
        self,
        debug: bool | None = None,
        dev: bool | None = None,
        dev_port: int | None = None,
        fastapi_kwargs: dict = {},
    ) -> None:
        if debug is None:
            self.debug = os.environ.get("PYNM_DEBUG", "False").lower() == "true"
        if dev is None:
            self.dev = os.environ.get("PYNM_DEV", "False").lower() == "true"
        if dev_port is None:
            self.dev_port = os.environ.get("PYNM_DEV_PORT", str(DEV_SERVER_PORT))

        super().__init__(
            title="PyNeuromodulation",
            description="PyNeuromodulation FastAPI backend",
            version=importlib.metadata.version("py_neuromodulation"),
            debug=self.debug,
            **fastapi_kwargs,
        )

        # Use the FastAPI logger for the backend
        self.logger = logging.getLogger("uvicorn.error")
        self.logger.warning(PYNM_DIR)

        if self.dev:
            cors_origins = (
                ["http://localhost:" + str(self.dev_port)] if self.dev else []
            )
            # Configure CORS
            self.add_middleware(
                CORSMiddleware,
                allow_origins=cors_origins,
                allow_credentials=True,
                allow_methods=["*"],
                allow_headers=["*"],
            )

        # Has to be before mounting static files
        self.setup_routes()

        # Serve static files
        if not self.dev:
            self.logger.info("Serving static files")
            self.mount(
                "/",
                StaticFiles(directory=PYNM_DIR / "gui" / "frontend", html=True),
                name="static",
            )

        self.websocket_manager = WebsocketManager()
        self.pynm_state = app_pynm.PyNMState()

    def setup_routes(self):
        @self.get("/api/health")
        async def healthcheck():
            return {"message": "API is working"}

        ####################
        ##### SETTINGS #####
        ####################
        @self.get("/api/settings")
        async def get_settings(
            reset: bool = Query(False, description="Reset settings to default"),
        ):
            if reset:
                settings = NMSettings.get_default()
            else:
                settings = self.pynm_state.stream.settings

            return settings.serialize_with_metadata()

        @self.post("/api/settings")
        async def update_settings(data: dict, validate_only: bool = Query(False)):
            try:
                # First, validate with Pydantic
                try:
                    # TODO: check if this works properly or needs model_validate_strings
                    validated_settings = NMSettings.model_validate(data)
                except ValidationError as e:
                    self.logger.error(f"Error validating settings: {e}")
                    if not validate_only:
                        # If validation failed but we wanted to upload, return error
                        raise HTTPException(
                            status_code=422,
                            detail={
                                "error": "Error validating settings",
                                "details": str(e),
                            },
                        )
                    # Else return list of errors
                    return {
                        "valid": False,
                        "errors": [err for err in e.errors()],
                        "details": str(e),
                    }

                # If validation succesful, return or update settings
                if validate_only:
                    return {
                        "valid": True,
                        "settings": validated_settings.serialize_with_metadata(),
                    }

                self.pynm_state.stream.settings = validated_settings
                self.logger.info("Settings successfully updated")

                return {
                    "valid": True,
                    "settings": self.pynm_state.stream.settings.serialize_with_metadata(),
                }

            # If something else than validation went wrong, return error
            except Exception as e:
                self.logger.error(f"Error validating/updating settings: {e}")
                raise HTTPException(
                    status_code=422,
                    detail={"error": "Error uploading settings", "details": str(e)},
                )

        ########################
        ##### PYNM CONTROL #####
        ########################

        @self.post("/api/stream-control")
        async def handle_stream_control(data: dict):
            action = data["action"]
            if action == "start":
                # TODO: create out_dir and experiment_name text filds in frontend
                self.logger.info("websocket:")
                self.logger.info(self.websocket_manager)
                self.logger.info("Starting stream")

                self.pynm_state.start_run_function(
                    websocket_manager=self.websocket_manager,
                )

            if action == "stop":
                self.logger.info("Stopping stream")
                self.pynm_state.stop_run_function()

            return {"message": f"Stream action '{action}' executed"}

        ####################
        ##### CHANNELS #####
        ####################

        @self.get("/api/channels")
        async def get_channels():
            channels = self.pynm_state.stream.channels
            self.logger.info(f"Sending channels: {channels}")
            if isinstance(channels, pd.DataFrame):
                return {"channels": channels.to_dict(orient="records")}
            else:
                raise HTTPException(
                    status_code=422,
                    detail={"error": "Channels is not a DataFrame"},
                )

        @self.post("/api/channels")
        async def update_channels(data: dict):
            try:
                new_channels = pd.DataFrame(data["channels"])
                self.logger.info(f"Received channels:\n {new_channels}")
                self.pynm_state.stream.channels = new_channels
                return {
                    "channels": self.pynm_state.stream.channels.to_dict(
                        orient="records"
                    )
                }
            except Exception as e:
                raise HTTPException(
                    status_code=422,
                    detail={"error": "Error updating channels", "details": str(e)},
                )

        ###################
        ### LSL STREAMS ###
        ###################

        @self.get("/api/LSL-streams")
        async def get_lsl_streams():
            from mne_lsl.lsl import resolve_streams

            return {
                "message": [
                    {
                        "dtype":  # MNE-LSL might return a class, so we get the name
                        getattr(stream.dtype, "__name__", str(stream.dtype)),
                        "name": stream.name,
                        "n_channels": stream.n_channels,
                        "sfreq": stream.sfreq,
                        "source_id": stream.source_id,
                        "stype": stream.stype,  # Stream type (e.g. EEG)
                        "created_at": stream.created_at,
                        "hostname": stream.hostname,
                        "session_id": stream.session_id,
                        "uid": stream.uid,
                        "protocol_version": stream.protocol_version,
                    }
                    for stream in resolve_streams()
                ]
            }

        @self.post("/api/setup-LSL-stream")
        async def setup_lsl_stream(data: dict):
            try:
                stream_name = data["stream_name"]
                self.logger.info(f"Attempting to setup LSL stream: '{stream_name}'")
                self.pynm_state.setup_lsl_stream(
                    lsl_stream_name=stream_name,
                    line_noise=data["line_noise"],
                )
                return {"message": f"LSL stream '{stream_name}' setup successfully"}
            except Exception as e:
                return {
                    "message": "LSL stream could not be setup",
                    "error": str(e),
                }

        @self.post("/api/setup-Offline-stream")
        async def setup_offline_stream(data: dict):
            self.logger.info("Data received to setup offline stream:")
            self.logger.info(data)
            try:
                self.pynm_state.setup_offline_stream(
                    file_path=data["file_path"],
                    line_noise=float(data["line_noise"]),
                )
                return {"message": "Offline stream setup successfully"}
            except ValueError:
                return {"message": "Offline stream could not be setup"}

        @self.post("/api/set-stream-params")
        async def set_stream_params(data: dict):
            try:
                self.pynm_state.stream.line_noise = float(data["line_noise"])
                self.pynm_state.stream.sfreq = float(data["sampling_rate"])
                self.pynm_state.experiment_name = data["experiment_name"]
                self.pynm_state.out_dir = data["out_dir"]
                self.pynm_state.decoding_model_path = data["decoding_path"]

                return {"message": "Stream parameters updated successfully"}
            except ValueError:
                return {"message": "Stream parameters could not be updated"}

        #######################
        ### PYNM ABOUT INFO ###
        #######################

        @self.get("/api/app-info")
        # TODO: fix this function
        async def get_app_info():
            metadata = importlib.metadata.metadata("py_neuromodulation")
            url_list = metadata.get_all("Project-URL")
            urls = (
                {url.split(",")[0]: url.split(",")[1] for url in url_list}
                if url_list
                else {}
            )

            classifier_list = metadata.get_all("Classifier")
            classifiers = (
                {
                    item[: item.find("::") - 1]: item[item.find("::") + 3 :]
                    for item in classifier_list
                }
                if classifier_list
                else {}
            )
            if "License" in classifiers:
                classifiers["License"] = classifiers["License"].split("::")[1]

            return {
                "version": metadata.get("Version", ""),
                "website": urls.get("Homepage", ""),
                "authors": [metadata.get("Author-email", "")],
                "maintainers": [metadata.get("Maintainer", "")],
                "repository": urls.get("Repository", ""),
                "documentation": urls.get("Documentation", ""),
                "license": classifiers["License"],
                # "launchMode": "debug" if app.debug else "release",
            }

        ##############################
        ### FILE BROWSER ENDPOINTS ###
        ##############################
        # Get home directory for the current user
        @self.get("/api/home_directory")
        async def home_directory():
            try:
                home_dir = str(Path.home())
                return {"home_directory": home_dir}
            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))

        # Get PYNM_DIR
        @self.get("/api/pynm_dir")
        async def get_pynm_dir():
            try:
                return {"pynm_dir": PYNM_DIR}
            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))

        # Get list of available drives in Windows systems
        @self.get("/api/drives")
        async def list_drives():
            if os.name == "nt":
                import string

                drives = []
                for letter in string.ascii_uppercase:
                    if Path(f"{letter}:").exists():
                        drives.append(f"{letter}:")

                return {"drives": drives}
            else:
                return {"drives": ["/"]}  # Unix-like systems have a single root

        # Get list of files and directories in a directory
        @self.get("/api/files")
        async def list_files(
            path: str = Query(default="", description="Directory path to list"),
            allowed_extensions: str = Query(
                default=",".join(ALLOWED_EXTENSIONS),
                description="Comma-separated list of allowed file extensions",
            ),
            show_hidden: bool = Query(
                default=False,
                description="Whether to show hidden files and directories",
            ),
        ) -> list[FileInfo]:
            try:
                if not path:
                    path = str(Path.home())

                if not Path(path).is_dir():
                    raise FileNotFoundError("The specified path is not a directory")

                allowed_ext = allowed_extensions.split(",")

                files = []
                for entry in Path(path).iterdir():
                    # Skip hidden files/directories if show_hidden is False
                    if not show_hidden and is_hidden(entry):
                        continue

                    if entry.is_file() and not any(
                        entry.name.lower().endswith(ext.lower()) for ext in allowed_ext
                    ):
                        continue

                    stats = entry.stat()
                    files.append(
                        FileInfo(
                            name=entry.name,
                            path=str(entry),
                            dir=str(entry.parent),
                            is_directory=entry.is_dir(),
                            size=stats.st_size if not entry.is_dir() else 0,
                            created_at=datetime.fromtimestamp(stats.st_birthtime),
                            modified_at=datetime.fromtimestamp(stats.st_mtime),
                        )
                    )
                return files
            except FileNotFoundError:
                raise HTTPException(status_code=404, detail="Directory not found")
            except PermissionError:
                raise HTTPException(status_code=403, detail="Permission denied")
            except Exception as e:
                raise HTTPException(status_code=500, detail=str(e))

        @self.get("/api/quick-access")
        def quick_access():
            return get_quick_access()

        ###########################
        ### WEBSOCKET ENDPOINTS ###
        ###########################
        @self.websocket("/ws")
        async def websocket_endpoint(websocket: WebSocket):
            await self.websocket_manager.connect(websocket)
            while True:
                try:
                    await websocket.receive_text()
                except Exception:
                    break



import numpy as np
import pandas as pd
import pytest

import py_neuromodulation as nm


def get_example_settings(test_arr: np.ndarray) -> tuple[nm.NMSettings, pd.DataFrame]:
    settings = nm.NMSettings.get_fast_compute()

    channels = nm.utils.get_default_channels_from_data(test_arr)

    return settings, channels


def test_different_sampling_rate_100Hz():
    sampling_rate_features = 100

    arr_test = np.random.random([2, 1020])
    settings, channels = get_example_settings(arr_test)

    settings.sampling_rate_features_hz = sampling_rate_features
    stream = nm.Stream(sfreq=1000, channels=channels, settings=settings, verbose=True)

    df = stream.run(
        arr_test,
        out_dir="./test_data",
        experiment_name="test_different_sampling_rate_100Hz",
    )

    # check the difference between time points
    assert np.diff(df["time"].iloc[:2]) / 1000 == (1 / sampling_rate_features)


def test_different_sampling_rate_10Hz():
    sampling_rate_features = 10

    arr_test = np.random.random([2, 1200])
    settings, channels = get_example_settings(arr_test)

    settings.sampling_rate_features_hz = sampling_rate_features
    stream = nm.Stream(sfreq=1000, channels=channels, settings=settings, verbose=True)

    df = stream.run(
        arr_test,
        out_dir="./test_data",
        experiment_name="test_different_sampling_rate_10Hz",
    )

    # check the difference between time points

    assert np.diff(df["time"].iloc[:2]) / 1000 == (1 / sampling_rate_features)


def test_different_sampling_rate_1Hz():
    sampling_rate_features = 1

    arr_test = np.random.random([2, 3000])
    settings, channels = get_example_settings(arr_test)

    settings.sampling_rate_features_hz = sampling_rate_features
    stream = nm.Stream(sfreq=1000, channels=channels, settings=settings, verbose=True)

    df = stream.run(
        arr_test,
        out_dir="./test_data",
        experiment_name="test_different_sampling_rate_1Hz",
    )

    # check the difference between time points

    assert np.diff(df["time"].iloc[:2]) / 1000 == (1 / sampling_rate_features)


def test_different_sampling_rate_0DOT1Hz():
    sampling_rate_features = 0.1

    arr_test = np.random.random([2, 30000])
    settings, channels = get_example_settings(arr_test)

    settings.sampling_rate_features_hz = sampling_rate_features
    stream = nm.Stream(sfreq=1000, channels=channels, settings=settings, verbose=True)

    df = stream.run(
        arr_test,
        out_dir="./test_data",
        experiment_name="test_different_sampling_rate_0DOT1Hz",
    )

    # check the difference between time points

    assert np.diff(df["time"].iloc[:2]) / 1000 == (1 / sampling_rate_features)


def test_wrong_initalization_of_segment_length_features_ms_and_osc_window_length():
    arr_test = np.random.random([2, 1200])
    settings, channels = get_example_settings(arr_test)

    settings.segment_length_features_ms = 800
    settings.fft_settings.windowlength_ms = 1000

    with pytest.raises(Exception):
        nm.Stream(sfreq=1000, channels=channels, settings=settings, verbose=True)


def test_different_segment_lengths():
    segment_length_features_ms = 800

    arr_test = np.random.random([2, 1200])
    settings, channels = get_example_settings(arr_test)

    settings.segment_length_features_ms = segment_length_features_ms
    settings.fft_settings.windowlength_ms = segment_length_features_ms

    stream = nm.Stream(sfreq=1000, channels=channels, settings=settings, verbose=True)

    df_seglength_800 = stream.run(
        arr_test,
        out_dir="./test_data",
        experiment_name="test_different_segment_lengths_800",
    )

    segment_length_features_ms = 1000

    arr_test = np.random.random([2, 1200])
    settings, channels = get_example_settings(arr_test)

    settings.segment_length_features_ms = segment_length_features_ms
    settings.fft_settings.windowlength_ms = segment_length_features_ms

    stream = nm.Stream(sfreq=1000, channels=channels, settings=settings, verbose=True)

    df_seglength_1000 = stream.run(
        arr_test,
        out_dir="./test_data",
        experiment_name="test_different_segment_lengths_1000",
    )
    # check the difference between time points

    print(df_seglength_1000.columns)
    assert (
        df_seglength_1000.iloc[0]["ch0_avgref_fft_theta_mean"]
        != df_seglength_800.iloc[0]["ch0_avgref_fft_theta_mean"]
    )



"""
Lab Streaming Layer (LSL) Example
=================================

This toolbox implements the lsl ecosystem which can be utilized for offline use cases as well as live streamings
In this example the data introduced in the first demo is being analyzed
in a similar manner, This time however integrating an lsl stream.

"""

# %%
from matplotlib import pyplot as plt
import py_neuromodulation as nm

# %%
# Let’s get the example data from the provided BIDS dataset and create the channels DataFrame.

if __name__ == "__main__":
    (
        RUN_NAME,
        PATH_RUN,
        PATH_BIDS,
        PATH_OUT,
        datatype,
    ) = nm.io.get_paths_example_data()

    (
        raw,
        data,
        sfreq,
        line_noise,
        coord_list,
        coord_names,
    ) = nm.io.read_BIDS_data(PATH_RUN=PATH_RUN)

    channels = nm.utils.set_channels(
        ch_names=raw.ch_names,
        ch_types=raw.get_channel_types(),
        reference="default",
        bads=raw.info["bads"],
        new_names="default",
        used_types=("ecog", "dbs", "seeg"),
        target_keywords=["MOV_RIGHT"],
    )

    # %%
    # Playing the Data
    # ----------------
    #
    # Now we need our data to be represeted in the LSL stream.
    # For this example an mne_lsl.Player is utilized, which is playing our earlier
    # recorded data. However, you could make use of any LSL source (live or
    # offline).
    # If you want to bind your own data source, make sure to specify the
    # necessary parameters (data type, type, name) accordingly.
    # If you are unsure about the parameters of your data source you can
    # always search for available lsl streams.
    #

    settings = nm.NMSettings.get_fast_compute()

    player = nm.stream.LSLOfflinePlayer(raw=raw, stream_name="example_stream")

    player.start_player(chunk_size=30, n_repeat=1)
    import time

    time.sleep(5)
    # %%
    # Creating the LSLStream object
    # -----------------------------
    #
    # Next let’s create a Stream analog to the First Demo’s example However as
    # we run the stream, we will set the *lsl-stream* value to True and pass
    # the stream name we earlier declared when initializing the player object.

    settings.features.welch = False
    settings.features.fft = True
    settings.features.bursts = False
    settings.features.sharpwave_analysis = False
    settings.features.coherence = False

    # %%
    stream = nm.Stream(
        sfreq=sfreq,
        channels=channels,
        settings=settings,
        coord_list=coord_list,
        verbose=True,
        line_noise=line_noise,
    )
    # %%
    # We then simply have to set the `stream_lsl` parameter to be `True` and specify the `stream_lsl_name`.
    # Schedule "stop" insertion in 5 seconds
    features = stream.run(
        is_stream_lsl=True,
        stream_lsl_name="example_stream",
        out_dir=PATH_OUT,
        experiment_name=RUN_NAME,
    )

    player.stop_player()

    # %%
    # We can then look at the computed features and check if the streamed data was processed correctly.
    # This can be verified by the time label:

    plt.plot(features.time, features.MOV_RIGHT)

    ######################################################################
    # Feature Analysis of Movement
    # ----------------------------
    # We can now check the movement averaged features of an ECoG channel.
    # Note that the path was here adapted to be documentation build compliant.

    feature_reader = nm.analysis.FeatureReader(
        feature_dir=PATH_OUT, feature_file=RUN_NAME
    )
    feature_reader.label_name = "MOV_RIGHT"
    feature_reader.label = feature_reader.feature_arr["MOV_RIGHT"]

    feature_reader.plot_target_averaged_channel(
        ch="ECOG_RIGHT_0",
        list_feature_keywords=None,
        epoch_len=4,
        threshold=0.5,
        ytick_labelsize=7,
        figsize_x=12,
        figsize_y=12,
    )



from collections.abc import Sequence
from collections import defaultdict
from itertools import product

from pydantic import model_validator
from typing import TYPE_CHECKING, Any, Callable

import numpy as np

if np.__version__ >= "2.0.0":
    from numpy._core._methods import _mean as np_mean  # type: ignore
else:
    from numpy.core._methods import _mean as np_mean

from py_neuromodulation.utils.types import (
    NMFeature,
    NMBaseModel,
    BoolSelector,
    FrequencyRange,
)

if TYPE_CHECKING:
    from py_neuromodulation import NMSettings

# Using low-level numpy mean function for performance, could do the same for the other estimators
ESTIMATOR_DICT = {
    "mean": np_mean,
    "median": np.median,
    "max": np.max,
    "min": np.min,
    "var": np.var,
}


class PeakDetectionSettings(NMBaseModel):
    estimate: bool = True
    distance_troughs_ms: float = 10
    distance_peaks_ms: float = 5


class SharpwaveFeatures(BoolSelector):
    peak_left: bool = False
    peak_right: bool = False
    num_peaks: bool = False
    trough: bool = False
    width: bool = False
    prominence: bool = True
    interval: bool = True
    decay_time: bool = False
    rise_time: bool = False
    sharpness: bool = True
    rise_steepness: bool = False
    decay_steepness: bool = False
    slope_ratio: bool = False


class SharpwaveEstimators(NMBaseModel):
    mean: list[str] = ["interval"]
    median: list[str] = []
    max: list[str] = ["prominence", "sharpness"]
    min: list[str] = []
    var: list[str] = []

    def keys(self):
        return ["mean", "median", "max", "min", "var"]

    def values(self):
        return [self.mean, self.median, self.max, self.min, self.var]


class SharpwaveSettings(NMBaseModel):
    sharpwave_features: SharpwaveFeatures = SharpwaveFeatures()
    filter_ranges_hz: list[FrequencyRange] = [
        FrequencyRange(5, 80),
        FrequencyRange(5, 30),
    ]
    detect_troughs: PeakDetectionSettings = PeakDetectionSettings()
    detect_peaks: PeakDetectionSettings = PeakDetectionSettings()
    estimator: SharpwaveEstimators = SharpwaveEstimators()
    apply_estimator_between_peaks_and_troughs: bool = True

    def disable_all_features(self):
        self.sharpwave_features.disable_all()
        for est in self.estimator.keys():
            self.estimator[est] = []

    @model_validator(mode="after")
    def test_settings(cls, settings):
        # check if all features are also enabled via an estimator
        estimator_list = [est for list_ in settings.estimator.values() for est in list_]

        for used_feature in settings.sharpwave_features.get_enabled():
            assert (
                used_feature in estimator_list
            ), f"Add estimator key for {used_feature}"

        return settings


class SharpwaveAnalyzer(NMFeature):
    def __init__(
        self, settings: "NMSettings", ch_names: Sequence[str], sfreq: float
    ) -> None:
        self.sw_settings = settings.sharpwave_analysis_settings
        self.sfreq = sfreq
        self.ch_names = ch_names
        self.list_filter: list[tuple[str, Any]] = []
        self.trough: list = []
        self.troughs_idx: list = []

        settings.validate()

        # FrequencyRange's are already ensured to have high > low
        # Test that the higher frequency is smaller than the sampling frequency
        for filter_range in settings.sharpwave_analysis_settings.filter_ranges_hz:
            assert filter_range[1] < sfreq, (
                "Filter range has to be smaller than sfreq, "
                f"got sfreq {sfreq} and filter range {filter_range}"
            )

        for filter_range in settings.sharpwave_analysis_settings.filter_ranges_hz:
            # Test settings
            # TODO: handle None values
            if filter_range[0] is None:
                self.list_filter.append(("no_filter", None))
            else:
                from mne.filter import create_filter

                self.list_filter.append(
                    (
                        f"range_{filter_range[0]:.0f}_{filter_range[1]:.0f}",
                        create_filter(
                            None,
                            sfreq,
                            l_freq=filter_range[0],
                            h_freq=filter_range[1],
                            fir_design="firwin",
                            # l_trans_bandwidth=None,
                            # h_trans_bandwidth=None,
                            # filter_length=str(sfreq) + "ms",
                            verbose=False,
                        ),
                    )
                )

        self.filter_names = [name for name, _ in self.list_filter]
        self.filters = np.vstack([filter for _, filter in self.list_filter])
        self.filters = np.tile(self.filters[None, :, :], (len(self.ch_names), 1, 1))

        self.used_features = self.sw_settings.sharpwave_features.get_enabled()

        # initializing estimator functions, respecitive for all sharpwave features
        self.estimator_dict: dict[str, dict[str, Callable]] = {
            feat: {
                est: ESTIMATOR_DICT[est]
                for est in self.sw_settings.estimator.keys()
                if feat in self.sw_settings.estimator[est]
            }
            for feat_list in self.sw_settings.estimator.values()
            for feat in feat_list
        }

        estimator_combinations = [
            (feature_name, estimator_name, estimator)
            for feature_name in self.used_features
            for estimator_name, estimator in self.estimator_dict[feature_name].items()
        ]

        filter_combinations = list(
            product(
                enumerate(self.ch_names), enumerate(self.filter_names), [False, True]
            )
        )

        self.estimator_key_map: dict[str, Callable] = {}
        self.combinations = []
        for (ch_idx, ch_name), (
            filter_idx,
            filter_name,
        ), detect_troughs in filter_combinations:
            for feature_name, estimator_name, estimator in estimator_combinations:
                key_name = f"{ch_name}_Sharpwave_{estimator_name.title()}_{feature_name}_{filter_name}"
                self.estimator_key_map[key_name] = estimator
            self.combinations.append(
                (
                    (ch_idx, ch_name),
                    (filter_idx, filter_name),
                    detect_troughs,
                    estimator_combinations,
                )
            )

        # Check required feature computations according to settings
        self.need_peak_left = (
            self.sw_settings.sharpwave_features.peak_left
            or self.sw_settings.sharpwave_features.prominence
        )
        self.need_peak_right = (
            self.sw_settings.sharpwave_features.peak_right
            or self.sw_settings.sharpwave_features.prominence
        )
        self.need_trough = (
            self.sw_settings.sharpwave_features.trough
            or self.sw_settings.sharpwave_features.prominence
        )

        self.need_decay_steepness = (
            self.sw_settings.sharpwave_features.decay_steepness
            or self.sw_settings.sharpwave_features.slope_ratio
        )

        self.need_rise_steepness = (
            self.sw_settings.sharpwave_features.rise_steepness
            or self.sw_settings.sharpwave_features.slope_ratio
        )

        self.need_steepness = self.need_rise_steepness or self.need_decay_steepness

    def calc_feature(self, data: np.ndarray) -> dict:
        """Given a new data batch, the peaks, troughs and sharpwave features
        are estimated. Importantly only new data is being analyzed here. In
        steps of 1/settings["sampling_rate_features] analyzed and returned.
        Pre-initialized filters are applied to each channel.

        Parameters
        ----------
        data (np.ndarray): 2d data array with shape [num_channels, samples]
        feature_results (dict): Features.py estimated features

        Returns
        -------
        feature_results (dict): set features for Features.py object
        """
        dict_ch_features: dict[str, dict[str, float]] = defaultdict(lambda: {})

        from scipy.signal import fftconvolve

        data = np.tile(data[:, None, :], (1, len(self.list_filter), 1))
        data = fftconvolve(data, self.filters, axes=2, mode="same")

        self.filtered_data = (
            data  # TONI: Expose filtered data for example 3, need a better way
        )

        feature_results = {}

        for (
            (ch_idx, ch_name),
            (filter_idx, filter_name),
            detect_troughs,
            estimator_combinations,
        ) in self.combinations:
            sub_data = data[ch_idx, filter_idx, :]

            key_name_pt = "Trough" if detect_troughs else "Peak"

            if (not detect_troughs and not self.sw_settings.detect_peaks.estimate) or (
                detect_troughs and not self.sw_settings.detect_troughs.estimate
            ):
                continue

            # the detect_troughs loop start with peaks, s.t. data does not need to be flipped
            sub_data = -sub_data if detect_troughs else sub_data
            # sub_data *= 1 - 2 * detect_troughs # branchless version

            waveform_results = self.analyze_waveform(sub_data)

            # for each feature take the respective fun.
            for feature_name, estimator_name, estimator in estimator_combinations:
                if feature_name == "num_peaks":
                    key_name = f"{ch_name}_Sharpwave_{feature_name}_{filter_name}"
                    if len(waveform_results[feature_name]) == 1:
                        dict_ch_features[key_name][key_name_pt] = waveform_results[feature_name][0]
                        continue
                    else:
                        raise ValueError("num_peaks should be a list with length 1")
                    # there can be only one num_peak in each batch
                feature_data = waveform_results[feature_name]
                key_name = f"{ch_name}_Sharpwave_{estimator_name.title()}_{feature_name}_{filter_name}"

                # zero check because no peaks can be also detected
                feature_data = estimator(feature_data) if len(feature_data) != 0 else 0
                dict_ch_features[key_name][key_name_pt] = feature_data

        if self.sw_settings.apply_estimator_between_peaks_and_troughs:
            # apply between 'Trough' and 'Peak' the respective function again
            # save only the 'est_fun' (e.g. max) between them

            # the key_name stays, since the estimator function stays between peaks and troughs
            for key_name, estimator in self.estimator_key_map.items():
                if len(dict_ch_features[key_name]) == 0:
                    # might happen if num_peaks was written in estimator
                    # e.g. estimator["mean"] = ["num_peaks"]
                    # for conveniance this doesn't raise an exception
                    continue

                feature_results[key_name] = estimator(
                    [
                        list(dict_ch_features[key_name].values())[0],
                        list(dict_ch_features[key_name].values())[1],
                    ]
                )
            # add here also the num_peaks features
            if self.sw_settings.sharpwave_features.num_peaks:
                for ch_name in self.ch_names:
                    for filter_name in self.filter_names:
                        key_name = f"{ch_name}_Sharpwave_num_peaks_{filter_name}"
                        feature_results[key_name] = np_mean([dict_ch_features[key_name]["Peak"],
                                                            dict_ch_features[key_name]["Trough"]])
        else:
            # otherwise, save all write all "flattened" key value pairs in feature_results
            for key, subdict in dict_ch_features.items():
                for key_sub, value_sub in subdict.items():
                    feature_results[key + "_analyze_" + key_sub] = value_sub

        return feature_results

    def analyze_waveform(self, data) -> dict:
        """Given the scipy.signal.find_peaks trough/peak distance
        settings specified sharpwave features are estimated.
        """

        from scipy.signal import find_peaks

        # TODO: find peaks is actually not that big a performance hit, but the rest
        # of this function is. Perhaps find_peaks can be put in a loop and the rest optimized somehow?
        peak_idx: np.ndarray = find_peaks(
            data, distance=self.sw_settings.detect_troughs.distance_peaks_ms
        )[0]
        trough_idx: np.ndarray = find_peaks(
            -data, distance=self.sw_settings.detect_troughs.distance_troughs_ms
        )[0]

        """ Find left and right peak indexes for each trough """
        peak_pointer = first_valid = last_valid = 0
        peak_idx_left_list: list[int] = []
        peak_idx_right_list: list[int] = []

        for i in range(len(trough_idx)):
            # Locate peak right of current trough
            while (
                peak_pointer < peak_idx.size and peak_idx[peak_pointer] < trough_idx[i]
            ):
                peak_pointer += 1

            if peak_pointer - 1 < 0:
                # If trough has no peak to it's left, it's not valid
                first_valid = i + 1  # Try with next one
                continue

            if peak_pointer == peak_idx.size:
                # If we went past the end of the peaks list, trough had no peak to its right
                continue

            last_valid = i
            peak_idx_left_list.append(peak_idx[peak_pointer - 1])
            peak_idx_right_list.append(peak_idx[peak_pointer])

        # Remove non valid troughs and make array of left and right peaks for  each trough
        trough_idx = trough_idx[first_valid : last_valid + 1]
        peak_idx_left = np.array(peak_idx_left_list, dtype=int)
        peak_idx_right = np.array(peak_idx_right_list, dtype=int)

        """ Calculate features (vectorized) """
        results: dict = {}

        if self.need_peak_left:
            results["peak_left"] = data[peak_idx_left]

        if self.need_peak_right:
            results["peak_right"] = data[peak_idx_right]

        if self.need_trough:
            results["trough"] = data[trough_idx]

        if self.sw_settings.sharpwave_features.interval:
            results["interval"] = np.concatenate((np.zeros(1), np.diff(trough_idx))) * (
                1000 / self.sfreq
            )

        if self.sw_settings.sharpwave_features.sharpness:
            # sharpess is calculated on a +- 5 ms window
            # valid troughs need 5 ms of margin on both sides
            troughs_valid = trough_idx[
                np.logical_and(
                    trough_idx - int(5 * (1000 / self.sfreq)) > 0,
                    trough_idx + int(5 * (1000 / self.sfreq)) < data.shape[0],
                )
            ]
            trough_height = data[troughs_valid]
            left_height = data[troughs_valid - int(5 * (1000 / self.sfreq))]
            right_height = data[troughs_valid + int(5 * (1000 / self.sfreq))]
            # results["sharpness"] = ((trough_height - left_height) + (trough_height - right_height)) / 2
            results["sharpness"] = trough_height - 0.5 * (left_height + right_height)

        if self.sw_settings.sharpwave_features.num_peaks:
            results["num_peaks"] = [
                trough_idx.shape[0]
            ]  # keep list to the estimator can be applied

        if self.need_steepness:
            # steepness is calculated as the first derivative
            steepness: np.ndarray = np.concatenate((np.zeros(1), np.diff(data)))

            # Create an array with the rise and decay steepness for each trough
            # 0th dimension for rise/decay, 1st for trough index, 2nd for timepoint
            steepness_troughs = np.zeros((2, trough_idx.shape[0], steepness.shape[0]))
            if self.need_rise_steepness or self.need_decay_steepness:
                for i in range(len(trough_idx)):
                    steepness_troughs[
                        0, i, 0 : trough_idx[i] - peak_idx_left[i] + 1
                    ] = steepness[peak_idx_left[i] : trough_idx[i] + 1]
                    steepness_troughs[
                        1, i, 0 : peak_idx_right[i] - trough_idx[i] + 1
                    ] = steepness[trough_idx[i] : peak_idx_right[i] + 1]

            if self.need_rise_steepness:
                # left peak -> trough
                # + 1 due to python syntax, s.t. the last element is included
                results["rise_steepness"] = np.max(
                    np.abs(steepness_troughs[0, :, :]), axis=1
                )

            if self.need_decay_steepness:
                # trough -> right peak
                results["decay_steepness"] = np.max(
                    np.abs(steepness_troughs[1, :, :]), axis=1
                )

            if self.sw_settings.sharpwave_features.slope_ratio:
                results["slope_ratio"] = (
                    results["rise_steepness"] - results["decay_steepness"]
                )

        if self.sw_settings.sharpwave_features.prominence:
            results["prominence"] = np.abs(
                (results["peak_right"] + results["peak_left"]) / 2 - results["trough"]
            )

        if self.sw_settings.sharpwave_features.decay_time:
            results["decay_time"] = (peak_idx_left - trough_idx) * (
                1000 / self.sfreq
            )  # ms

        if self.sw_settings.sharpwave_features.rise_time:
            results["rise_time"] = (peak_idx_right - trough_idx) * (
                1000 / self.sfreq
            )  # ms

        if self.sw_settings.sharpwave_features.width:
            results["width"] = peak_idx_right - peak_idx_left  # ms

        return results



import py_neuromodulation as nm

import numpy as np


def test_setting_computation_time():
    """Intantiate test for feature computation with fixed time duration.
    The number of output features should match the ratio of the sampling rate of the data and raw signal sampling rate.
    """

    data_duration_s = 5
    sampling_rate_features_hz = 200
    fs = 1000
    data = np.random.random((1, int(data_duration_s * fs)))

    settings = nm.NMSettings.get_fast_compute()
    settings.segment_length_features_ms = 1000  # start afte 1 second
    settings.features.fft = False
    settings.features.raw_hjorth = True
    stream = nm.Stream(
        sfreq=fs,
        data=data,
        sampling_rate_features_hz=sampling_rate_features_hz,
        settings=settings,
    )

    features = stream.run(
        out_dir="./test_data", experiment_name="test_setting_computation_time"
    )

    # test if features up till the last sample was computed
    assert features.time.iloc[-1] == data_duration_s * fs


    # test that the time difference between two samples is the feature sampling rate
    assert (
        features.time.iloc[1] - features.time.iloc[0]
    ) == fs / sampling_rate_features_hz

    assert features.time.iloc[0] == settings.segment_length_features_ms


def test_float_fs():
    """Change sampling rate here to be float, s.t. rounding issues are not affecting overall number of
    computed features.
    """

    data_duration_s = 5
    sampling_rate_features_hz = 200
    fs = 1111.111
    data = np.random.random((1, int(data_duration_s * fs)))

    settings = nm.NMSettings.get_fast_compute()
    settings.segment_length_features_ms = 333  # start after 1 second

    settings.features.fft = False
    settings.features.raw_hjorth = True
    stream = nm.Stream(
        sfreq=fs,
        data=data,
        sampling_rate_features_hz=sampling_rate_features_hz,
        settings=settings,
    )

    features = stream.run(out_dir="./test_data", experiment_name="test_float_fs")

    # test that the time difference between two samples is the feature sampling rate
    assert (
        features.time.iloc[1] - features.time.iloc[0]
    ) == 1000 / sampling_rate_features_hz

    # the timing of the first sample cannot be directly inferred from the segment length
    # the samples are computed based on rounding to full indices of the original data



from pathlib import Path
from py_neuromodulation.utils.types import _PathLike
import logging

INFOFORMAT = "%(name)s:\t%(message)s"
DEBUGFORMAT = "%(asctime)s:%(levelname)s:%(name)s:%(filename)s:%(funcName)s:%(lineno)d:\t%(message)s"

LOG_LEVELS = {
    "DEBUG": (logging.DEBUG, DEBUGFORMAT),
    "INFO": (logging.INFO, INFOFORMAT),
    "WARNING": (logging.WARN, DEBUGFORMAT),
    "ERROR": (logging.ERROR, DEBUGFORMAT),
}


class NMLogger(logging.Logger):
    """
    Subclass of logging.Logger with some extra functionality
    """

    def __init__(self, name: str, level: str = "INFO") -> None:
        super().__init__(name, LOG_LEVELS[level][0])

        self.setLevel(level)

        self._console_handler = logging.StreamHandler()
        self._console_handler.setLevel(level)
        self._console_handler.setFormatter(logging.Formatter(LOG_LEVELS[level][1]))

        self.addHandler(self._console_handler)

    def set_level(self, level: str):
        """
        Set console logging level
        """
        self.setLevel(level)
        self._console_handler.setLevel(level)
        self._console_handler.setFormatter(logging.Formatter(LOG_LEVELS[level][1]))

    def log_to_file(self, path: _PathLike, mode: str = "w"):
        """
        Add file handlers to the logger

        Parameters
        ----------
        path: directory where to save logfiles
        mode : str, ('w', 'a')
            w: overwrite files
            a: append to files
        """

        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)

        self.debug_file_handler = logging.FileHandler(path / "logfile_pydebug.log")
        self.debug_file_handler.setLevel(logging.DEBUG)
        self.debug_file_handler.setFormatter(logging.Formatter(DEBUGFORMAT))

        self.info_file_handler = logging.FileHandler(
            path / "logfile_pyinfo.log", mode=mode
        )
        self.info_file_handler.setLevel(logging.INFO)
        self.info_file_handler.setFormatter(logging.Formatter(INFOFORMAT))

        self.addHandler(self.info_file_handler)
        self.addHandler(self.debug_file_handler)



import asyncio
import sys
from typing import Callable

if sys.platform.startswith("win"):
    import msvcrt
else:
    import termios
    import tty


class KeyboardListener:
    def __init__(self, event_callback: tuple[str, Callable] | None = None):
        self.callbacks = {}
        self.running = False

        if event_callback is not None:
            self.on_press(*event_callback)

    def on_press(self, key, callback):
        self.callbacks[key] = callback

    async def _windows_listener(self):
        while self.running:
            if msvcrt.kbhit():
                key = msvcrt.getch().decode("utf-8").lower()
                if key in self.callbacks:
                    await self.callbacks[key]()
            await asyncio.sleep(0.01)

    async def _unix_listener(self):
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            while self.running:
                key = sys.stdin.read(1).lower()
                if key in self.callbacks:
                    await self.callbacks[key]()
                await asyncio.sleep(0.01)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)

    async def start(self):
        self.running = True
        if sys.platform.startswith("win"):
            await self._windows_listener()
        else:
            await self._unix_listener()

    def stop(self):
        self.running = False



import numpy as np
from mne_connectivity import make_signals_in_freq_bands

import py_neuromodulation as nm


def test_coherence():
    """Check that coherence features compute properly and match expected values."""
    # Simulate connectivity data (interaction at specified frequency band)
    sfreq = 500  # Hz
    n_epochs = 1
    n_times = sfreq * 10  # samples
    fband = (15, 20)  # frequency band of interaction, Hz
    trans = 2  # transition bandwidth of signal, Hz
    delay = 50  # samples
    epochs = make_signals_in_freq_bands(
        n_seeds=1,
        n_targets=1,
        freq_band=fband,
        n_epochs=n_epochs,
        n_times=n_times,
        sfreq=sfreq,
        trans_bandwidth=trans,
        connection_delay=delay,
        ch_names=["seed", "target"],
        snr=0.9,  # change here requires change in `signal/noise_con` vars below
        rng_seed=44
    )

    # Set up py_nm channels info
    ch_names = epochs.ch_names
    ch_types = epochs.get_channel_types()
    channels = nm.utils.set_channels(
        ch_names=ch_names,
        ch_types=ch_types,
        reference="default",
        bads=None,
        new_names="default",
        used_types=tuple(np.unique(ch_types)),
        target_keywords=None,
    )

    # Set up pn_nm processing settings
    settings = nm.NMSettings.get_default()
    settings.reset()
    settings.features.coherence = True

    # redefine freq. bands of interest
    # (accounts for signal-noise transition bandwdith when defining frequencies)
    settings.frequency_ranges_hz = {
        "signal": {  # strong connectivity expected
            "frequency_low_hz": fband[0],
            "frequency_high_hz": fband[1],
        },
        "noise_low": {  # weak connectivity expected from 1 Hz to start of interaction
            "frequency_low_hz": 1,
            "frequency_high_hz": fband[0] - trans * 2,
        },
        "noise_high": {  # weak connectivity expected from end of interaction to Nyquist
            "frequency_low_hz": fband[1] + trans * 2,
            "frequency_high_hz": sfreq // 2 - 1,
        },
    }
    settings.coherence_settings.frequency_bands = ["signal", "noise_low", "noise_high"]
    settings.coherence_settings.nperseg = sfreq
    # only average within each band required
    settings.coherence_settings.features = {
        "mean_fband": True, "max_fband": False, "max_allfbands": False
    }

    # unique all-to-all connectivity indices, i.e.: [[0, 1]]
    settings.coherence_settings.channels = [ch_names]

    # do not normalise features for this test!
    # (normalisation changes interpretability of connectivity values, making it harder to
    # define 'expected' connectivity values)
    settings.postprocessing.feature_normalization = False

    # Set up py_nm stream
    stream = nm.Stream(
        settings=settings,
        channels=channels,
        path_grids=None,
        verbose=True,
        sfreq=epochs.info["sfreq"],
    )

    # Compute connectivity
    features = stream.run(
        epochs.get_data(copy=False)[0],  # extract first (and only) epoch from obj
        out_dir="./test_data",
        experiment_name="test_coherence",
    )

    # Aggregate results over windows
    results = {key: None for key in features.keys()}
    results.pop("time")
    for key in results.keys():
        # average over windows; take absolute before averaging icoh values
        results[key] = np.abs(features[key].values).mean()

    node_name = "seed_to_target"
    #for con_method in ["coh", "icoh"]:
    con_method = "icoh"
    # Assert that frequencies of simulated interaction have strong connectivity
    assert (
        results[f"{con_method}_{node_name}_mean_fband_signal"] >
        results[f"{con_method}_{node_name}_mean_fband_noise_low"]
    )

    assert (
        results[f"{con_method}_{node_name}_mean_fband_signal"] >
        results[f"{con_method}_{node_name}_mean_fband_noise_high"]
    )



import numpy as np
from typing import TYPE_CHECKING


from py_neuromodulation.utils.types import NMBaseModel
from py_neuromodulation.utils.pydantic_extensions import NMErrorList


if TYPE_CHECKING:
    from py_neuromodulation.stream.settings import NMSettings


class KalmanSettings(NMBaseModel):
    Tp: float = 0.1
    sigma_w: float = 0.7
    sigma_v: float = 1.0
    frequency_bands: list[str] = [
        "theta",
        "alpha",
        "low_beta",
        "high_beta",
        "low_gamma",
        "high_gamma",
        "HFA",
    ]

    def validate_fbands(self, settings: "NMSettings") -> NMErrorList:
        errors: NMErrorList = NMErrorList()

        if not all(
            (item in settings.frequency_ranges_hz for item in self.frequency_bands)
        ):
            errors.add_error(
                "Frequency bands for Kalman filter must also be specified in "
                "frequency_ranges_hz.",
                location=[
                    "kalman_filter_settings",
                    "frequency_bands",
                ],
            )

        return errors


def define_KF(Tp, sigma_w, sigma_v):
    """Define Kalman filter according to white noise acceleration model.
    See DOI: 10.1109/TBME.2009.2038990  for explanation
    See https://filterpy.readthedocs.io/en/latest/kalman/KalmanFilter.html#r64ca38088676-2 for implementation details

    Parameters
    ----------
    Tp : float
        prediction interval
    sigma_w : float
        process noise
    sigma_v : float
        measurement noise

    Returns
    -------
    filterpy.KalmanFilter
        initialized KalmanFilter object
    """
    from .kalman_filter_external import KalmanFilter

    f = KalmanFilter(dim_x=2, dim_z=1)
    f.x = np.array([0, 1])  # x here sensor signal and it's first derivative
    f.F = np.array([[1, Tp], [0, 1]])
    f.H = np.array([[1, 0]])
    f.R = sigma_v
    f.Q = np.array(
        [
            [(sigma_w**2) * (Tp**3) / 3, (sigma_w**2) * (Tp**2) / 2],
            [(sigma_w**2) * (Tp**2) / 2, (sigma_w**2) * Tp],
        ]
    )
    f.P = np.cov([[1, 0], [0, 1]])
    return f



from typing import Any
import py_neuromodulation as nm
import multiprocessing as mp


class StreamBackendInterface:
    """Handles stream data output via queues"""

    def __init__(
        self, feature_queue: mp.Queue, raw_data_queue: mp.Queue, control_queue: mp.Queue
    ):
        self.feature_queue = feature_queue
        self.rawdata_queue = raw_data_queue
        self.control_queue = control_queue

    def send_command(self, command: str) -> None:
        """Send a command through the control queue"""
        try:
            self.control_queue.put(command)
        except Exception as e:
            nm.logger.error(f"Error sending command: {e}")

    def send_features(self, features: dict[str, Any]) -> None:
        """Send feature data through the feature queue"""
        try:
            nm.logger.debug("backend_interface.send_features before feature put in queue")
            self.feature_queue.put(features)
            nm.logger.debug("backend_interface.send_features after feature put in queue")
        except Exception as e:
            nm.logger.error(f"Error sending features: {e}")

    def send_raw_data(self, data: dict[str, Any]) -> None:
        """Send raw data through the rawdata queue"""
        try:
            self.rawdata_queue.put(data)
        except Exception as e:
            nm.logger.error(f"Error sending raw data: {e}")

    def check_control_signals(self) -> str | None:
        """Check for control signals (non-blocking)"""
        try:
            if not self.control_queue.empty():
                return self.control_queue.get_nowait()
            return None
        except Exception as e:
            nm.logger.error(f"Error checking control signals: {e}")
            return None



"""
R-Map computation
=================

"""

# %%
# Across patient decoding using R-Map optimal connectivity
# --------------------------------------------------------
#
# ECoG electrode placement is commonly very heterogeneous across patients and cohorts.
# To still facilitate approaches that are able to perform decoding applications without patient individual training,
# two across-patient decoding approaches were previously investigated for movement decoding:
#
#
# * grid-point decoding
# * optimal connectivity channel decoding
#
#
# First, the grid-point decoding approach relies on definition of a cortical or subcortical grid.
# Data from individual grid points is then interpolated onto those common grid points.
# The approach was also explained in the :doc:`plot_4_example_gridPointProjection` notebook.
#
# .. image:: ../_static/RMAP_figure.png
#     :alt: R-Map and grid point approach for decoding without patient-individual training
#
# The R-Map decoding approach relies on the other hand on computation of whole brain connectivity. The electrode MNI space locations need to be known,
# then the following steps can be performed for decoding without patient individual training:
#
# #. Using the `wjn_toolbox <https://github.com/neuromodulation/wjn_toolbox>`_ *wjn_specrical_roi* function, the MNI coordinates can be transformed into NIFTI (.nii) files, containing the electrode contact region of interest (ROI):
#
#    .. code-block:: python
#
#      wjn_spherical_roi(roiname, mni, 4)
#
# #. For the given *ROI.nii* files, the LeadDBS `LeadMapper <https://netstim.gitbook.io/leaddbs/connectomics/lead-mapper>`_ tool can be used for functional or structural connectivity estimation.
#
# #. The py_neuromodulation :class:`~RMAP.py` module can then compute the R-Map given the contact-individual connectivity fingerprints:
#
#    .. code-block:: python
#
#        RMAP.calculate_RMap_numba(fingerprints, performances)
#
# #. The fingerprints from test-set patients can then be correlated with the calculated R-Map:
#
#    .. code-block:: python
#
#       RMAP.get_corr_numba(fp, fp_test)
#
# #. The channel with highest correlation can then be selected for decoding without individual training. :class:`~RMAP.py` contain already leave one channel and leave one patient out cross validation functions:
#
#    .. code-block:: python
#
#       RMAP.leave_one_sub_out_cv(l_fps_names, l_fps_dat, l_per, sub_list)
#
# #. The obtained R-Map correlations can then be estimated statistically and plotted against true correlates:
#
#    .. code-block:: python
#
#       RMAP.plot_performance_prediction_correlation(per_left_out, per_predict, out_path_save)
#
#
# sphinx_gallery_thumbnail_path = '_static/RMAP_figure.png'



class PARRMArtifactRejection:
    """
    This module enables training of a PARRM filter before computation,
    that can in real-time then be applied.
    https://pyparrm.readthedocs.io/en/stable/
    """
    def __init__(self, data, sampling_freq, artefact_freq, verbose=False):
        from pyparrm import PARRM

        self.data = data
        self.sampling_freq = sampling_freq
        self.artefact_freq = artefact_freq
        self.verbose = verbose

        self.parrm = PARRM(
            data=data,
            sampling_freq=sampling_freq,
            artefact_freq=artefact_freq,
            verbose=False,
        )

    def filter_data(self):
        self.parrm.find_period()
        self.parrm.create_filter(
            filter_direction="both",
        )
        filtered_data = self.parrm.filter_data()

        return filtered_data



"""
Reference:  B Hjorth
            EEG analysis based on time domain properties
            Electroencephalogr Clin Neurophysiol. 1970 Sep;29(3):306-10.
            DOI: 10.1016/0013-4694(70)90143-4
"""

from typing import TYPE_CHECKING
import numpy as np
from collections.abc import Sequence

from py_neuromodulation.utils.types import NMFeature

if TYPE_CHECKING:
    from py_neuromodulation import NMSettings


class Hjorth(NMFeature):
    def __init__(
        self, settings: "NMSettings", ch_names: Sequence[str], sfreq: float
    ) -> None:
        self.ch_names = ch_names

    def calc_feature(self, data: np.ndarray) -> dict:
        var = np.var(data, axis=-1)
        deriv1 = np.diff(data, axis=-1)
        deriv2 = np.diff(deriv1, axis=-1)
        deriv1_var = np.var(deriv1, axis=-1)
        deriv2_var = np.var(deriv2, axis=-1)
        deriv1_mobility = np.sqrt(deriv2_var / deriv1_var)

        activity = np.nan_to_num(var)
        mobility = np.nan_to_num(np.sqrt(deriv1_var / var))
        complexity = np.nan_to_num(deriv1_mobility / mobility)

        feature_results = {}
        for ch_idx, ch_name in enumerate(self.ch_names):
            feature_results[f"{ch_name}_RawHjorth_Activity"] = activity[ch_idx]
            feature_results[f"{ch_name}_RawHjorth_Mobility"] = mobility[ch_idx]
            feature_results[f"{ch_name}_RawHjorth_Complexity"] = complexity[ch_idx]

        return feature_results


class Raw(NMFeature):
    def __init__(
        self, settings: "NMSettings", ch_names: Sequence[str], sfreq: float
    ) -> None:
        self.ch_names = ch_names

    def calc_feature(self, data: np.ndarray) -> dict:
        feature_results = {}

        for ch_idx, ch_name in enumerate(self.ch_names):
            feature_results["_".join([ch_name, "raw"])] = data[ch_idx, -1]

        return feature_results



import numpy as np
from typing import cast

from py_neuromodulation.utils.types import NMPreprocessor
from py_neuromodulation import logger


class NotchFilter(NMPreprocessor):
    def __init__(
        self,
        sfreq: float,
        line_noise: float | None = None,
        freqs: np.ndarray | None = None,
        notch_widths: int | np.ndarray | None = 3,
        trans_bandwidth: float = 6.8,
    ) -> None:
        from mne.filter import create_filter

        if line_noise is None and freqs is None:
            raise ValueError(
                "Either line_noise or freqs must be defined if notch_filter is"
                "activated."
            )

        if freqs is None:
            freqs = np.arange(line_noise, sfreq / 2, line_noise, dtype=int)

        if freqs.size > 0 and freqs[-1] >= sfreq / 2:
            freqs = freqs[:-1]

        # Code is copied from filter.py notch_filter
        if freqs.size == 0:
            self.filter_bank = None
            logger.warning(
                "WARNING: notch_filter is activated but data is not being"
                " filtered. This may be due to a low sampling frequency or"
                " incorrect specifications. Make sure your settings are"
                f" correct. Got: {sfreq = }, {line_noise = }, {freqs = }."
            )
            return

        filter_length = int(sfreq - 1)
        if notch_widths is None:
            notch_widths = freqs / 200.0
        elif np.any(notch_widths < 0):
            raise ValueError("notch_widths must be >= 0")
        else:
            notch_widths = np.atleast_1d(notch_widths)
            if len(notch_widths) == 1:
                notch_widths = notch_widths[0] * np.ones_like(freqs)
            elif len(notch_widths) != len(freqs):
                raise ValueError(
                    "notch_widths must be None, scalar, or the " "same length as freqs"
                )
        notch_widths = cast(np.ndarray, notch_widths)  # For MyPy only, no runtime cost

        # Speed this up by computing the fourier coefficients once
        tb_half = trans_bandwidth / 2.0
        lows = [freq - nw / 2.0 - tb_half for freq, nw in zip(freqs, notch_widths)]
        highs = [freq + nw / 2.0 + tb_half for freq, nw in zip(freqs, notch_widths)]

        self.filter_bank = create_filter(
            data=None,
            sfreq=sfreq,
            l_freq=highs,
            h_freq=lows,
            filter_length=filter_length,  # type: ignore
            l_trans_bandwidth=tb_half,  # type: ignore
            h_trans_bandwidth=tb_half,  # type: ignore
            method="fir",
            iir_params=None,
            phase="zero",
            fir_window="hamming",
            fir_design="firwin",
            verbose=False,
        )

    def process(self, data: np.ndarray) -> np.ndarray:
        if self.filter_bank is None:
            return data

        from mne.filter import _overlap_add_filter

        return _overlap_add_filter(
            x=data,
            h=self.filter_bank,
            n_fft=None,
            phase="zero",
            picks=None,
            n_jobs=1,
            copy=True,
            pad="reflect_limited",
        )



import nibabel as nib
import numpy as np
import scipy.io as sio
import os


def write_connectome_mat(
    PATH_Fingerprints: str = r"D:\Connectome_RMAP_OUT\ROIs\HCP1000 6K",
    PATH_CONNECTOME: str = os.path.join(
        "py_neuromodulation",
        "ConnectivityDecoding",
        "connectome_struct.mat",
    ),
    func_: bool = False,
):
    # connectome = sio.loadmat(PATH_CONNECTOME)  # check if read was successful

    # load all fingerprints and put them in .npy
    dict_connectome = {}
    if func_ is False:
        files_fps = [f for f in os.listdir(PATH_Fingerprints) if ".nii" in f]
    else:
        files_fps = [
            f for f in os.listdir(PATH_Fingerprints) if "func_seed_AvgR_Fz.nii" in f
        ]

    # I except 1025 files, check which ones are missing
    missing_files = []

    for i in range(1, 1026):
        MISSING = False

        if func_ is False:
            if f"ROI-{i}_struc_seed.nii" not in files_fps:
                missing_files.append(f"ROI-{i}_struc_seed.nii")
                MISSING = True
        else:
            if f"ROI-{i}_func_seed_AvgR_Fz.nii" not in files_fps:
                missing_files.append(f"ROI-{i}_func_seed_AvgR_Fz.nii")
                MISSING = True

        if MISSING:
            ROI_file = os.path.join(
                r"D:\Connectome_RMAP_OUT\whole_brain\ROIs", f"ROI-{i}.nii"
            )
            # copy the ROI file to the following folder:
            PATH_ROI_OUT = r"D:\Connectome_RMAP_OUT\whole_brain\ROI_missing_struc"
            import shutil

            shutil.copy(ROI_file, os.path.join(PATH_ROI_OUT, f"ROI-{i}.nii"))

    for idx, f in enumerate(files_fps):
        # load the .nii file and put it all in in a dictionary with the name of the file
        fp = nib.load(os.path.join(PATH_Fingerprints, f)).get_fdata().astype(np.float16)
        if "struc" in f:
            dict_connectome[f[f.find("ROI-") + 4 : f.find("_struc")]] = fp
        else:
            dict_connectome[
                f[f.find("ROI-") + 4 : f.find("_func_seed_AvgR_Fz.nii")]
            ] = fp
    # save the dictionary
    sio.savemat(
        PATH_CONNECTOME,
        dict_connectome,
    )


if __name__ == "__main__":
    write_connectome_mat(
        PATH_Fingerprints=r"D:\Connectome_RMAP_OUT\whole_brain\struc\HCP1000 6K",
        PATH_CONNECTOME=os.path.join(
            "py_neuromodulation",
            "ConnectivityDecoding",
            "connectome_whole_brain_struc.mat",
        ),
    )  # 58 files are missing

    write_connectome_mat(
        PATH_Fingerprints=r"D:\Connectome_RMAP_OUT\whole_brain\func",
        PATH_CONNECTOME=os.path.join(
            "py_neuromodulation",
            "ConnectivityDecoding",
            "connectome_whole_brain_func.mat",
        ),
        func_=True,
    )

    write_connectome_mat(
        PATH_Fingerprints=r"D:\Connectome_RMAP_OUT\hull\func\GSP 1000 (Yeo 2011)_Full Set (Yeo 2011)",
        PATH_CONNECTOME=os.path.join(
            "py_neuromodulation",
            "ConnectivityDecoding",
            "connectome_hull_func.mat",
        ),
        func_=True,
    )  # all there

    write_connectome_mat(
        PATH_Fingerprints=r"D:\Connectome_RMAP_OUT\hull\struc\HCP1000 6K",
        PATH_CONNECTOME=os.path.join(
            "py_neuromodulation",
            "ConnectivityDecoding",
            "connectome_hull_struc.mat",
        ),
    )  # 5 missing



"""Module for handling channels."""

from collections.abc import Iterable
from typing import TYPE_CHECKING
import numpy as np

if TYPE_CHECKING:
    import pandas as pd

_LFP_TYPES = ["seeg", "dbs", "lfp"]  # must be lower-case


def set_channels(
    ch_names: list[str],
    ch_types: list[str],
    reference: list | str = "default",
    bads: list[str] | None = None,
    new_names: str | list[str] = "default",
    ecog_only: bool = False,
    used_types: Iterable[str] | None = ("ecog", "dbs", "seeg"),
    target_keywords: Iterable[str] | None = ("mov", "squared", "label"),
) -> "pd.DataFrame":
    """Return dataframe with channel-specific settings in channels format.

    Return an channels dataframe with the columns: "name", "rereference",
    "used", "target", "type", "status", "new_name"]. "name" is set to ch_names,
    "rereference" can be specified individually. "used" is set to 1 for all
    channel types specified in `used_types`, else to 0. "target" is set to 1
    for all channels containing any of the `target_keywords`, else to 0.

    Possible channel types:
    https://github.com/mne-tools/mne-python/blob/6ae3b22033c745cce5cd5de9b92da54c13c36484/doc/_includes/channel_types.rst

    Arguments
    ---------
        ch_names : list
            list of channel names.
        ch_types : list
            list of channel types. Should optimally be of the types: "ECOG",
            "DBS" or "SEEG".
        reference : str | list of str | None, default: 'default'
            re-referencing scheme. Default is "default". This sets ECOG channel
            references to "average" and creates a bipolar referencing scheme
            for LFP/DBS/SEEG channels, where each channel is referenced to
            the adjacent lower channel, split by left and right hemisphere.
            For this, the channel names must contain the substring "_L_" and/or
            "_R_" (lower or upper case). CAVE: Adjacent channels will be
            determined using the sort() function.
        bads :  str | list of str, default: None
            channels that should be marked as bad and not be used for
            average re-referencing etc.
        new_names : list of str | None, default: 'default'
            new channel names that should be used when writing out the
            features and results. Useful when applying re-referencing. Set to
            'None' if no renaming should be performed. 'default' will infer
            channel renaming from re-referencing information. If a list is
            given, it should be in the same order as 'ch_names'.
        ecog_only : boolean, default: False
            if True, set only 'ecog' channel type to used
        used_types : iterable of str | None, default : ("ecog", "dbs", "seeg")
            data channel types to be used. Set to `None` to use no channel
            types.
        target_keywords : iterable of str | None, default : ("ecog", "dbs", "seeg")
            keywords for target channels

    Returns
    -------
        df: DataFrame in channels format
    """
    import pandas as pd

    if not (len(ch_names) == len(ch_types)):
        raise ValueError(
            "Number of `ch_names` and `ch_types` must match."
            f"Got: {len(ch_names)} `ch_names` and {len(ch_types)} `ch_types`."
        )

    df = pd.DataFrame(
        data=None,
        columns=[
            "name",
            "rereference",
            "used",
            "target",
            "type",
            "status",
            "new_name",
        ],
    )
    df["name"] = ch_names

    if used_types:
        if isinstance(used_types, str):
            used_types = [
                used_types
            ]  # Even if the user passes only ("ecog"), the if statement bellow will work
        used_list = []
        for ch_type in ch_types:
            if any(use_type.lower() == ch_type.lower() for use_type in used_types):
                used_list.append(1)
            else:
                used_list.append(0)
        df["used"] = used_list
    else:
        df["used"] = 0

    if target_keywords:
        if isinstance(target_keywords, str):
            target_keywords = [target_keywords]
        targets = []
        for ch_name in ch_names:
            if any(kw.lower() in ch_name.lower() for kw in target_keywords):
                targets.append(1)
            else:
                targets.append(0)
        df["target"] = targets
    else:
        df["target"] = 0

    # note: BIDS types are in caps, mne.io.RawArray types lower case
    # so that 'type' will be in lower case here
    df["type"] = ch_types

    if ecog_only:
        df.loc[(df["type"] == "seeg") | (df["type"] == "dbs"), "used"] = 0

    if isinstance(reference, str):
        if reference.lower() == "default":
            df = _get_default_references(df=df, ch_names=ch_names, ch_types=ch_types)
        else:
            raise ValueError(
                "`reference` must be either `default`, `None` or "
                "an iterable of new reference channel names. "
                f"Got: {reference}."
            )

    elif isinstance(reference, list):
        if len(ch_names) != len(reference):
            raise ValueError(
                "Number of `ch_names` and `reference` must match."
                f"Got: {len(ch_names)} `ch_names` and {len(reference)}"
                " `references`."
            )
        df["rereference"] = reference
    elif not reference:
        df.loc[:, "rereference"] = "None"
    else:
        raise ValueError(
            "`reference` must be either `default`, None or "
            "an iterable of new reference channel names. "
            f"Got: {reference}."
        )

    if bads:
        if isinstance(bads, str):
            bads = [bads]
        df["status"] = ["bad" if ch in bads else "good" for ch in ch_names]
        df.loc[df["status"] == "bad", "used"] = (
            0  # setting bad channels to not being used
        )
    else:
        df["status"] = "good"

    if not new_names:
        df["new_name"] = ch_names
    elif isinstance(new_names, str):
        if new_names.lower() != "default":
            raise ValueError(
                "`new_names` must be either `default`, None or "
                "an iterable of new channel names. Got: "
                f"{new_names}."
            )
        new_names = []
        for name, ref in zip(df["name"], df["rereference"]):
            if ref == "None":
                new_names.append(name)
            elif isinstance(ref, float):
                if np.isnan(ref):
                    new_names.append(name)
            elif ref == "average":
                new_names.append(name + "_avgref")
            else:
                new_names.append(name + "_" + ref)
        df["new_name"] = new_names
    elif hasattr(new_names, "__iter__"):
        if len(new_names) != len(ch_names):
            raise ValueError(
                "Number of `ch_names` and `new_names` must match."
                f" Got: {len(ch_names)} `ch_names` and {len(new_names)}"
                " `new_names`."
            )
        else:
            df["new_name"] = ch_names
    else:
        raise ValueError(
            "`new_names` must be either `default`, None or"
            f" an iterable of new channel names. Got: {new_names}."
        )
    return df


def _get_default_references(
    df: "pd.DataFrame", ch_names: list[str], ch_types: list[str]
) -> "pd.DataFrame":
    """Add references with default settings (ECOG CAR, LFP bipolar)."""
    ecog_chs = []
    lfp_chs = []
    other_chs = []
    for ch_name, ch_type in zip(ch_names, ch_types):
        if "ecog" in ch_type.lower() or "ecog" in ch_name.lower():
            ecog_chs.append(ch_name)
        elif any(
            lfp_type in ch_type.lower() or lfp_type in ch_name.lower()
            for lfp_type in _LFP_TYPES
        ):
            lfp_chs.append(ch_name)
        else:
            other_chs.append(ch_name)
    lfp_l = [
        lfp_ch
        for lfp_ch in lfp_chs
        if ("_l_" in lfp_ch.lower()) or ("_left_" in lfp_ch.lower())
    ]
    lfp_l.sort()
    lfp_r = [
        lfp_ch
        for lfp_ch in lfp_chs
        if ("_r_" in lfp_ch.lower()) or ("_right_" in lfp_ch.lower())
    ]
    lfp_r.sort()
    lfp_l_refs = [lfp_l[i - 1] if i > 0 else lfp_l[-1] for i, _ in enumerate(lfp_l)]
    lfp_r_refs = [lfp_r[i - 1] if i > 0 else lfp_r[-1] for i, _ in enumerate(lfp_r)]
    ref_idx = list(df.columns).index("rereference")
    if len(ecog_chs) > 1:
        for ecog_ch in ecog_chs:
            df.iloc[df[df["name"] == ecog_ch].index[0], ref_idx] = "average"
    if (
        len(lfp_l) > 1
    ):  # if there is only a single channel, the channel would be subtracted from itself
        for i, lfp in enumerate(lfp_l):
            df.iloc[df[df["name"] == lfp].index[0], ref_idx] = lfp_l_refs[i]
    if len(lfp_r) > 1:
        for i, lfp in enumerate(lfp_r):
            df.iloc[df[df["name"] == lfp].index[0], ref_idx] = lfp_r_refs[i]
    for other_ch in other_chs:
        df.iloc[df[df["name"] == other_ch].index[0], ref_idx] = "None"

    df = df.replace(np.nan, "None")

    return df


def get_default_channels_from_data(
    data: "np.ndarray | pd.DataFrame",
    car_rereferencing: bool = True,
):
    """Return default channels dataframe with
    ecog datatype, no bad channels, no targets, common average rereferencing

    Parameters
    ----------
    data : np.ndarray
        Data array in shape (n_channels, n_time)
    car_rereferencing : bool, optional
        use common average rereferencing, by default True

    Returns
    -------
    pd.DataFrame
        nm_channel dataframe containing columns:
         - name
         - rereference
         - used
         - target
         - type
         - status
         - new_name
    """
    import pandas as pd

    ch_name = [f"ch{idx}" for idx in range(data.shape[0])]
    status = ["good" for _ in range(data.shape[0])]
    type_nm = ["ecog" for _ in range(data.shape[0])]

    if car_rereferencing:
        rereference = ["average" for _ in range(data.shape[0])]
        new_name = [f"{ch}_avgref" for ch in ch_name]
    else:
        rereference = ["None" for _ in range(data.shape[0])]
        new_name = ch_name

    new_name = [f"{ch}_avgref" for ch in ch_name]
    target = np.array([0 for _ in range(data.shape[0])])
    used = np.array([1 for _ in range(data.shape[0])])

    df = pd.DataFrame()
    df["name"] = ch_name
    df["rereference"] = rereference
    df["used"] = used
    df["target"] = target
    df["type"] = type_nm
    df["status"] = status
    df["new_name"] = new_name

    return df



from collections.abc import Iterator
import time
from typing import TYPE_CHECKING
import numpy as np
from py_neuromodulation import logger
from mne_lsl.lsl import resolve_streams
import os

if TYPE_CHECKING:
    from py_neuromodulation import NMSettings


class LSLStream:
    """
    Class is used to create and connect to a LSL stream and pull data from it.
    """

    def __init__(self, settings: "NMSettings", stream_name: str | None = None) -> None:
        """
        Initialize the LSL stream.

        Parameters:
        -----------
        settings : settings.NMSettings object
        stream_name : str, optional
            Name of the stream to connect to. If not provided, the first available stream is used.

        Raises:
        -------
        RuntimeError
            If no stream is running under the provided name or if there are multiple streams running
            under the same name.
        """
        from mne_lsl.stream import StreamLSL

        self.stream: StreamLSL
        # self.keyboard_interrupt = False

        self.settings = settings
        self._n_seconds_wait_before_disconnect = 3
        try:
            if stream_name is None:
                stream_name = resolve_streams()[0].name
                logger.info(
                    f"Stream name not provided. Using first available stream: {stream_name}"
                )
            self.stream = StreamLSL(name=stream_name, bufsize=2).connect(timeout=2)
        except Exception as e:
            msg = f"Could not connect to stream: {e}. Either no stream is running under the name {stream_name} or there is several streams under this name."
            logger.exception(msg)
            raise RuntimeError(msg)

        if self.stream.sinfo is None:
            raise RuntimeError("Stream info is None. Check if the stream is running.")

        self.winsize = settings.segment_length_features_ms / self.stream.sinfo.sfreq
        self.sampling_interval = 1 / self.settings.sampling_rate_features_hz

        # If not running the generator when the escape key is pressed.
        self.headless: bool = not os.environ.get("DISPLAY")
        # if not self.headless:
        # from py_neuromodulation.utils.keyboard import KeyboardListener

        # self.listener = KeyboardListener(("esc", self.set_keyboard_interrupt))
        # self.listener.start()

    def get_next_batch(self) -> Iterator[tuple[np.ndarray, np.ndarray]]:
        self.last_time = time.time()
        check_data = None
        data = None
        stream_start_time = None

        while self.stream.connected:
            time_diff = time.time() - self.last_time  # in s
            time.sleep(0.005)
            if time_diff >= self.sampling_interval:
                self.last_time = time.time()

                logger.debug(f"Pull data - current time: {self.last_time}")
                logger.debug(f"time since last data pull {time_diff} seconds")

                if time_diff >= 2 * self.sampling_interval:
                    logger.warning(
                        "Feature computation time between two consecutive samples"
                        "was twice the feature sampling interval"
                    )
                if data is not None:
                    check_data = data

                data, timestamp = self.stream.get_data(winsize=self.winsize)
                if stream_start_time is None:
                    stream_start_time = timestamp[0]

                logger.info(f"Stream time: {timestamp[-1] - stream_start_time}")

                for i in range(self._n_seconds_wait_before_disconnect):
                    if (
                        data is not None
                        and check_data is not None
                        and np.allclose(data, check_data, atol=1e-7, rtol=1e-7)
                    ):
                        logger.warning(
                            f"No new data incoming. Disconnecting stream in {3 - i} seconds."
                        )
                        time.sleep(1)
                        i += 1
                        if i == self._n_seconds_wait_before_disconnect:
                            self.stream.disconnect()
                            logger.warning("Stream disconnected.")
                            break

                yield timestamp - stream_start_time, data.astype(float)
